#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------

""" Defines the stub functions used for creating concrete implementations of
    the standard EditorFactory subclasses supplied with the Traits package.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

import logging

from traits.api import HasPrivateTraits, TraitError
from traits.trait_base import ETSConfig

#-------------------------------------------------------------------------
#  Logging:
#-------------------------------------------------------------------------

logger = logging.getLogger(__name__)

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# List of implemented UI toolkits:
TraitUIToolkits = ['qt4', 'wx', 'null']

#-------------------------------------------------------------------------
#  Data:
#-------------------------------------------------------------------------

# The current GUI toolkit object being used:
_toolkit = None

#-------------------------------------------------------------------------
#  Low-level GUI toolkit selection function:
#-------------------------------------------------------------------------

try:
    provisional_toolkit = ETSConfig.provisional_toolkit
except AttributeError:
    from contextlib import contextmanager

    # for backward compatibility
    @contextmanager
    def provisional_toolkit(toolkit_name):
        """ Perform an operation with toolkit provisionally set

        This sets the toolkit attribute of the ETSConfig object set to the
        provided value. If the operation fails with an exception, the toolkit
        is reset to nothing.
        """
        if ETSConfig.toolkit:
            raise AttributeError("ETSConfig toolkit is already set")
        ETSConfig.toolkit = toolkit_name
        try:
            yield
        except:
            # reset the toolkit state
            ETSConfig._toolkit = ''
            raise


def _import_toolkit(name):
    return __import__(name, globals=globals(), level=1).toolkit


def assert_toolkit_import(name):
    """ Raise an error if a toolkit with the given name should not be allowed
    to be imported.
    """
    if ETSConfig.toolkit and ETSConfig.toolkit != name:
        raise RuntimeError("Importing from %s backend after selecting %s "
                           "backend!" % (name, ETSConfig.toolkit))


def toolkit_object(name, raise_exceptions=False):
    """ Return the toolkit specific object with the given name.  The name
    consists of the relative module path and the object name separated by a
    colon.
    """

    mname, oname = name.split(':')

    class Unimplemented(object):
        """ This is returned if an object isn't implemented by the selected
        toolkit.  It raises an exception if it is ever instantiated.
        """

        def __init__(self, *args, **kwargs):
            raise NotImplementedError(
                "The %s traits backend doesn't "
                "implement %s" %
                (ETSConfig.toolkit, oname))

    be_obj = Unimplemented
    be_mname = toolkit().__module__.split('.')[-2] + '.' + mname
    try:
        module = __import__(
            be_mname, globals=globals(), fromlist=[oname], level=1
        )
        try:
            be_obj = getattr(module, oname)
        except AttributeError as e:
            if raise_exceptions:
                raise e
    except ImportError as e:
        if raise_exceptions:
            raise e

    return be_obj


def toolkit(*toolkits):
    """ Selects and returns a low-level GUI toolkit.

    Use this function to get a reference to the current toolkit.

    Parameters
    ----------
    *toolkits : strings
        Toolkit names to try if toolkit not already selected.  If not supplied,
        defaults to order in TraitUIToolkits variable.

    Returns
    -------
    toolkit
        Appropriate concrete Toolkit subclass for selected toolkit.
    """
    global _toolkit

    # If _toolkit has already been set, simply return it.
    if _toolkit is not None:
        return _toolkit

    if ETSConfig.toolkit:
        # If a toolkit has already been set for ETSConfig, then use it:
        _toolkit = _import_toolkit(ETSConfig.toolkit)
        return _toolkit
    else:
        if not toolkits:
            toolkits = TraitUIToolkits

        for toolkit_name in toolkits:
            try:
                with provisional_toolkit(toolkit_name):
                    _toolkit = _import_toolkit(toolkit_name)
                    return _toolkit
            except (AttributeError, ImportError) as exc:
                # import failed, reset toolkit to none, log error and try again
                msg = "Could not import traits UI backend '{0}'"
                logger.info(msg.format(toolkit_name))
                if logger.getEffectiveLevel() <= logging.INFO:
                    logger.exception(exc)
        else:
            # Try using the null toolkit and printing a warning
            try:
                with provisional_toolkit('null'):
                    _toolkit = _import_toolkit('null')
                    import warnings
                    msg = (
                        "Unable to import the '{0}' backend for traits UI; " +
                        "using the 'null' backend instead.")
                    warnings.warn(msg.format(toolkit_name), RuntimeWarning)
                    return _toolkit

            except ImportError as exc:
                logger.exception(exc)
                raise TraitError("Could not import any UI toolkit. Tried:" +
                                 ', '.join(toolkits))

#-------------------------------------------------------------------------
#  'Toolkit' class (abstract base class):
#-------------------------------------------------------------------------


class Toolkit(HasPrivateTraits):
    """ Abstract base class for GUI toolkits.
    """

    #-------------------------------------------------------------------------
    #  Create GUI toolkit specific user interfaces using information from the
    #  specified UI object:
    #-------------------------------------------------------------------------

    def ui_panel(self, ui, parent):
        """ Creates a GUI-toolkit-specific panel-based user interface using
            information from the specified UI object.
        """
        raise NotImplementedError

    def ui_subpanel(self, ui, parent):
        """ Creates a GUI-toolkit-specific subpanel-based user interface using
            information from the specified UI object.
        """
        raise NotImplementedError

    def ui_livemodal(self, ui, parent):
        """ Creates a GUI-toolkit-specific modal "live update" dialog user
            interface using information from the specified UI object.
        """
        raise NotImplementedError

    def ui_live(self, ui, parent):
        """ Creates a GUI-toolkit-specific non-modal "live update" window user
            interface using information from the specified UI object.
        """
        raise NotImplementedError

    def ui_modal(self, ui, parent):
        """ Creates a GUI-toolkit-specific modal dialog user interface using
            information from the specified UI object.
        """
        raise NotImplementedError

    def ui_nonmodal(self, ui, parent):
        """ Creates a GUI-toolkit-specific non-modal dialog user interface using
            information from the specified UI object.
        """
        raise NotImplementedError

    def ui_popup(self, ui, parent):
        """ Creates a GUI-toolkit-specific temporary "live update" popup dialog
            user interface using information from the specified UI object.
        """
        raise NotImplementedError

    def ui_popover(self, ui, parent):
        """ Creates a GUI-toolkit-specific temporary "live update" popup dialog
            user interface using information from the specified UI object.
        """
        raise NotImplementedError

    def ui_info(self, ui, parent):
        """ Creates a GUI-toolkit-specific temporary "live update" popup dialog
            user interface using information from the specified UI object.
        """
        raise NotImplementedError

    def ui_wizard(self, ui, parent):
        """ Creates a GUI-toolkit-specific wizard dialog user interface using
            information from the specified UI object.
        """
        raise NotImplementedError

    def view_application(self, context, view, kind=None, handler=None,
                         id='', scrollable=None, args=None):
        """ Creates a GUI-toolkit-specific modal dialog user interface that
            runs as a complete application using information from the
            specified View object.

        Parameters
        ----------
        context : object or dictionary
            A single object or a dictionary of string/object pairs, whose trait
            attributes are to be edited. If not specified, the current object is
            used.
        view : view or string
            A View object that defines a user interface for editing trait
            attribute values.
        kind : string
            The type of user interface window to create. See the
            **traitsui.view.kind_trait** trait for values and
            their meanings. If *kind* is unspecified or None, the **kind**
            attribute of the View object is used.
        handler : Handler object
            A handler object used for event handling in the dialog box. If
            None, the default handler for Traits UI is used.
        id : string
            A unique ID for persisting preferences about this user interface,
            such as size and position. If not specified, no user preferences
            are saved.
        scrollable : Boolean
            Indicates whether the dialog box should be scrollable. When set to
            True, scroll bars appear on the dialog box if it is not large enough
            to display all of the items in the view at one time.

        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Positions the associated dialog window on the display:
    #-------------------------------------------------------------------------

    def position(self, ui):
        """ Positions the associated dialog window on the display.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Shows a 'Help' window for a specified UI and control:
    #-------------------------------------------------------------------------

    def show_help(self, ui, control):
        """ Shows a Help window for a specified UI and control.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Sets the title for the UI window:
    #-------------------------------------------------------------------------

    def set_title(self, ui):
        """ Sets the title for the UI window.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Sets the icon for the UI window:
    #-------------------------------------------------------------------------

    def set_icon(self, ui):
        """ Sets the icon for the UI window.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Saves user preference information associated with a UI window:
    #-------------------------------------------------------------------------

    def save_window(self, ui):
        """ Saves user preference information associated with a UI window.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Rebuilds a UI after a change to the content of the UI:
    #-------------------------------------------------------------------------

    def rebuild_ui(self, ui):
        """ Rebuilds a UI after a change to the content of the UI.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Converts a keystroke event into a corresponding key name:
    #-------------------------------------------------------------------------

    def key_event_to_name(self, event):
        """ Converts a keystroke event into a corresponding key name.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Hooks all specified events for all controls in a ui so that they can be
    #  routed to the corrent event handler:
    #-------------------------------------------------------------------------

    def hook_events(self, ui, control, events=None, handler=None):
        """ Hooks all specified events for all controls in a UI so that they
            can be routed to the correct event handler.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Routes a 'hooked' event to the corrent handler method:
    #-------------------------------------------------------------------------

    def route_event(self, ui, event):
        """ Routes a "hooked" event to the corrent handler method.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Indicates that an event should continue to be processed by the toolkit
    #-------------------------------------------------------------------------

    def skip_event(self, event):
        """ Indicates that an event should continue to be processed by the
            toolkit.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Destroys a specified GUI toolkit control:
    #-------------------------------------------------------------------------

    def destroy_control(self, control):
        """ Destroys a specified GUI toolkit control.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Destroys all of the child controls of a specified GUI toolkit control:
    #-------------------------------------------------------------------------

    def destroy_children(self, control):
        """ Destroys all of the child controls of a specified GUI toolkit
            control.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Returns a ( width, height ) tuple containing the size of a specified
    #  toolkit image:
    #-------------------------------------------------------------------------

    def image_size(self, image):
        """ Returns a ( width, height ) tuple containing the size of a
            specified toolkit image.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Returns a dictionary of useful constants:
    #-------------------------------------------------------------------------

    def constants(self):
        """ Returns a dictionary of useful constants.

            Currently, the dictionary should have the following key/value pairs:

            - WindowColor': the standard window background color in the toolkit
              specific color format.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Returns a renderer used to render 'themed' table cells for a specified
    #  TableColumn object:
    #-------------------------------------------------------------------------

    def themed_cell_renderer(self, column):
        """ Returns a renderer used to render 'themed' table cells for a
            specified TableColum object.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  GUI toolkit dependent trait definitions:
    #-------------------------------------------------------------------------

    def color_trait(self, *args, **traits):
        raise NotImplementedError

    def rgb_color_trait(self, *args, **traits):
        raise NotImplementedError

    def rgba_color_trait(self, *args, **traits):
        raise NotImplementedError

    def font_trait(self, *args, **traits):
        raise NotImplementedError

    def kiva_font_trait(self, *args, **traits):
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  'Editor' class methods:
    #-------------------------------------------------------------------------

    def ui_editor(self):
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  'EditorFactory' factory methods:
    #-------------------------------------------------------------------------

    def array_editor(self, *args, **traits):
        raise NotImplementedError

    def boolean_editor(self, *args, **traits):
        raise NotImplementedError

    def button_editor(self, *args, **traits):
        raise NotImplementedError

    def check_list_editor(self, *args, **traits):
        raise NotImplementedError

    def code_editor(self, *args, **traits):
        raise NotImplementedError

    def color_editor(self, *args, **traits):
        raise NotImplementedError

    def compound_editor(self, *args, **traits):
        raise NotImplementedError

    def custom_editor(self, *args, **traits):
        raise NotImplementedError

    def directory_editor(self, *args, **traits):
        raise NotImplementedError

    def drop_editor(self, *args, **traits):
        raise NotImplementedError

    def dnd_editor(self, *args, **traits):
        raise NotImplementedError

    def enum_editor(self, *args, **traits):
        raise NotImplementedError

    def file_editor(self, *args, **traits):
        raise NotImplementedError

    def font_editor(self, *args, **traits):
        raise NotImplementedError

    def key_binding_editor(self, *args, **traits):
        raise NotImplementedError

    def history_editor(self, *args, **traits):
        raise NotImplementedError

    def html_editor(self, *args, **traits):
        raise NotImplementedError

    def image_editor(self, *args, **traits):
        raise NotImplementedError

    def image_enum_editor(self, *args, **traits):
        raise NotImplementedError

    def instance_editor(self, *args, **traits):
        raise NotImplementedError

    def list_editor(self, *args, **traits):
        raise NotImplementedError

    def list_str_editor(self, *args, **traits):
        raise NotImplementedError

    def null_editor(self, *args, **traits):
        raise NotImplementedError

    def ordered_set_editor(self, *args, **traits):
        raise NotImplementedError

    def plot_editor(self, *args, **traits):
        raise NotImplementedError

    def range_editor(self, *args, **traits):
        raise NotImplementedError

    def rgb_color_editor(self, *args, **traits):
        raise NotImplementedError

    def rgba_color_editor(self, *args, **traits):
        raise NotImplementedError

    def shell_editor(self, *args, **traits):
        raise NotImplementedError

    def table_editor(self, *args, **traits):
        raise NotImplementedError

    def tabular_editor(self, *args, **traits):
        raise NotImplementedError

    def text_editor(self, *args, **traits):
        raise NotImplementedError

    def title_editor(self, *args, **traits):
        raise NotImplementedError

    def tree_editor(self, *args, **traits):
        raise NotImplementedError

    def tuple_editor(self, *args, **traits):
        raise NotImplementedError

    def value_editor(self, *args, **traits):
        raise NotImplementedError



""" A widget for displaying bitmapped images. """


# Enthought library imports.
from traits.api import Bool, Event, Instance, File, Interface, Unicode
from pyface.tasks.i_editor import IEditor


class IImageEditor(IEditor):
    """ A widget for editing text. """

    #### 'IPythonEditor' interface ############################################

    # Object being editor is a file
    obj = Instance(File)

    # The pathname of the file being edited.
    path = Unicode

    #### Events ####

    # The contents of the editor has changed.
    changed = Event

    ###########################################################################
    # 'IPythonEditor' interface.
    ###########################################################################

    def load(self, path=None):
        """ Loads the contents of the editor. """

    def save(self, path=None):
        """ Saves the contents of the editor. """



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2008 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#  Author: Duncan Child <duncan@enthought.com>
#
#-----------------------------------------------------------------------------
# The code for two-stage unpickling support has been taken from a PEP draft
# prepared by Dave Peterson and Prabhu Ramachandran.

""" An unpickler that is tolerant of class refactorings, and implements a
two-stage pickling process to make it possible to unpickle complicated Python
object hierarchies where the unserialized state of an object depends on the
state of other objects in the same pickle.
"""

# Standard library imports.
import sys
import logging
from os import path
from types import GeneratorType

if sys.version_info[0] >= 3:
    from pickle import _Unpickler as Unpickler
else:
    from pickle import Unpickler

from pickle import UnpicklingError, BUILD

# Enthought library imports
from traits.api import HasTraits, Instance


# Setup a logger for this module
logger = logging.getLogger(__name__)


##############################################################################
# constants
##############################################################################

# The name we backup the original setstate method to.
_BACKUP_NAME = '__enthought_sweet_pickle_original_setstate__'

# The name of the setstate method we hook
_SETSTATE_NAME = '__setstate__'

# The name we store our unpickling data under.
_UNPICKLER_DATA = '__enthought_sweet_pickle_unpickler__'


##############################################################################
# function '__replacement_setstate__'
##############################################################################

def __replacement_setstate__(self, state):
    """ Called to enable an unpickler to modify the state of this instance.
    """
    # Retrieve the unpickling information and use it to let the unpickler
    # modify our state.
    unpickler, module, name = getattr(self, _UNPICKLER_DATA)
    state = unpickler.modify_state(self, state, module, name)

    # If we were given a state, apply it to this instance now.
    if state is not None:

        # Save our state
        logger.debug('Final state: %s', state)
        self.__dict__.update(state)


##############################################################################
# function 'load_build_with_meta_data'
##############################################################################

def load_build_with_meta_data(self):
    """ Called prior to the actual load_build() unpickling method which primes
        the state dictionary with meta-data.
    """

    # Access the state object and check if it is a dictionary (state may also be
    # a tuple, which is used for other unpickling build operations).  Proceed to
    # the standard load_build() if the state obj is not a dict.
    state = self.stack[-1]
    if type(state) == dict:

        # If a file object is used, reference the file name
        if hasattr(self._file, 'name'):
            pickle_file_name = path.abspath(self._file.name)
        else :
            pickle_file_name = ""

        # Add any meta-data needed by __setstate__() methods here...
        state['_pickle_file_name'] = pickle_file_name

    # Call the standard load_build() method
    return self.load_build()


##############################################################################
# class 'NewUnpickler'
##############################################################################
class NewUnpickler(Unpickler):
    """ An unpickler that implements a two-stage pickling process to make it
    possible to unpickle complicated Python object hierarchies where the
    unserialized state of an object depends on the state of other objects in
    the same pickle.
    """

    def load(self, max_pass=-1):
        """Read a pickled object representation from the open file.

        Return the reconstituted object hierarchy specified in the file.
        """
        # List of objects to be unpickled.
        self.objects = []

        # We overload the load_build method.
        dispatch = self.dispatch
        dispatch[BUILD[0]] = NewUnpickler.load_build

        # call the super class' method.
        ret = Unpickler.load(self)
        self.initialize(max_pass)
        self.objects = []

        # Reset the Unpickler's dispatch table.
        dispatch[BUILD[0]] = Unpickler.load_build
        return ret

    def initialize(self, max_pass):
        # List of (object, generator) tuples that initialize objects.
        generators = []

        # Execute object's initialize to setup the generators.
        for obj in self.objects:
            if hasattr(obj, '__initialize__') and \
                   callable(obj.__initialize__):
                ret = obj.__initialize__()
                if isinstance(ret, GeneratorType):
                    generators.append((obj, ret))
                elif ret is not None:
                    raise UnpicklingError('Unexpected return value from '
                        '__initialize__.  %s returned %s' % (obj, ret))

        # Ensure a maximum number of passes
        if max_pass < 0:
            max_pass = len(generators)

        # Now run the generators.
        count = 0
        while len(generators) > 0:
            count += 1
            if count > max_pass:
                not_done = [x[0] for x in generators]
                msg = """Reached maximum pass count %s.  You may have
                         a deadlock!  The following objects are
                         uninitialized: %s""" % (max_pass, not_done)
                raise UnpicklingError(msg)
            for o, g in generators[:]:
                try:
                    g.next()
                except StopIteration:
                    generators.remove((o, g))

    # Make this a class method since dispatch is a class variable.
    # Otherwise, supposing the initial sweet_pickle.load call (which would
    # have overloaded the load_build method) makes a pickle.load call at some
    # point, we would have the dispatch still pointing to
    # NewPickler.load_build whereas the object being passed in will be an
    # Unpickler instance, causing a TypeError.
    @classmethod
    def load_build(cls, obj):
        # Just save the instance in the list of objects.
        if isinstance(obj, NewUnpickler):
            obj.objects.append(obj.stack[-2])
        Unpickler.load_build(obj)


##############################################################################
# class 'VersionedUnpickler'
##############################################################################

class VersionedUnpickler(NewUnpickler, HasTraits):
    """ An unpickler that is tolerant of class refactorings.

        This class reads in a pickled file and applies the transforms
        specified in its updater to generate a new hierarchy of objects
        which are at the current version of the classes they are instances
        of.

        Note that the creation of an updater is kept out of this class to
        ensure that the class can be reused in different situations.
        However, if no updater is provided during construction, then the
        global registry updater will be used.
    """

    ##########################################################################
    # Traits
    ##########################################################################

    ### public 'VersionedUnpickler' interface ################################

    # The updater used to modify the objects being unpickled.
    updater = Instance('apptools.sweet_pickle.updater.Updater')


    ##########################################################################
    # 'object' interface
    ##########################################################################

    ### operator methods #####################################################

    def __init__(self, file, **kws):
        super(VersionedUnpickler, self).__init__(file)

        self._file = file
        if self.updater is None:
            from global_registry import get_global_registry
            self.updater = get_global_registry()
        logger.debug('VersionedUnpickler [%s] using Updater [%s]', self,
            self.updater)

        # Update the BUILD instruction to use an overridden load_build method
        # NOTE: this is being disabled since, on some platforms, the object
        # is replaced with a regular Unpickler instance, creating a traceback:
        # AttributeError: Unpickler instance has no attribute '_file'
        # ...not sure how this happens since only a VersionedUnpickler has
        # the BUILD instruction replaced with one that uses _file, and it
        # should have _file defined.
        #self.dispatch[BUILD[0]] = load_build_with_meta_data


    ##########################################################################
    # 'Unpickler' interface
    ##########################################################################

    ### public interface #####################################################

    def find_class(self, module, name):
        """ Returns the class definition for the named class within the
            specified module.

            Overridden here to:

            - Allow updaters to redirect to a different class, possibly
              within a different module.
            - Ensure that any setstate hooks for the class are called
              when the instance of this class is unpickled.
        """
        # Remove any extraneous characters that an Unpickler might handle
        # but a user wouldn't have included in their mapping definitions.
        module = module.strip()
        name = name.strip()

        # Attempt to find the class, this may cause a new mapping for that
        # very class to be introduced.  That's why we ignore the result.
        try:
            klass = super(VersionedUnpickler, self).find_class(module, name)
        except:
            pass

        # Determine the target class that the requested class should be
        # mapped to according to our updater.  The target class is the one
        # at the end of any chain of mappings.
        original_module, original_name = module, name
        if self.updater is not None and \
            self.updater.has_class_mapping(module, name):
            module, name = self._get_target_class(module, name)
            if module != original_module or name != original_name:
                logger.debug('Unpickling [%s.%s] as [%s.%s]', original_module,
                    original_name, module, name)

        # Retrieve the target class definition
        try:
            klass = super(VersionedUnpickler, self).find_class(module, name)
        except Exception, e:
            from apptools.sweet_pickle import UnpicklingError
            logger.debug('Traceback when finding class [%s.%s]:' \
                         % (module, name), exc_info=True)
            raise UnpicklingError('Unable to load class [%s.%s]. '
                'Original exception was, "%s".  map:%s' % (
                module, name, str(e), self.updater.class_map))

        # Make sure we run the updater's state functions if any are declared
        # for the target class.
        if self.updater is not None \
            and self._has_state_function(original_module, original_name):
            self._add_unpickler(klass, original_module, original_name)

        return klass


    ##########################################################################
    # 'VersionedUnpickler' interface
    ##########################################################################

    ### public interface #####################################################

    def modify_state(self, obj, state, module, name):
        """ Called to update the specified state dictionary, which represents
            the class of the specified name within the specified module, to
            complete the unpickling of the specified object.
        """
        # Remove our setstate hook and associated data to ensure that
        # instances unpickled through some other framework don't call us.
        # IMPORTANT: Do this first to minimize the time this hook is in place!
        self._remove_unpickler(obj.__class__)

        # Determine what class and version we're starting from and going to.
        # If there is no version information, then assume version 0. (0 is
        # like an unversioned version.)
        source_key = self.updater.get_version_attribute(module, name)
        source_version = state.get(source_key, 0)
        target_key = self.updater.get_version_attribute(
            obj.__class__.__module__, obj.__class__.__name__)
        target_version = getattr(obj, target_key, 0)

        # Iterate through all the updates to the state by going one version
        # at a time.  Note that we assume there is exactly one path from our
        # starting class and version to our ending class and version.  As a
        # result, we assume we update a given class to its latest version
        # before looking for any class mappings.  Note that the version in the
        # updater is the version to convert *TO*.
        version = source_version
        next_version = version + 1
        while True:

            # Iterate through all version updates for the current class.
            key = self.updater.get_version_attribute(module, name)
            while (module, name, next_version) in self.updater.state_functions:
                functions = self.updater.state_functions[(module, name,
                    next_version)]
                for f in functions:
                    logger.debug('Modifying state from [%s.%s (v.%s)] to ' + \
                        '[%s.%s (v.%s)] using function %s', module, name,
                        version, module, name, next_version, f)
                    state = f(state)

                # Avoid infinite loops due to versions not changing.
                new_version = state.get(key, version)
                if new_version == version:
                    new_version = version + 1
                version = new_version
                next_version = version + 1

            # If there is one, move to the next class in the chain.  (We
            # explicitly keep the version number the same.)
            if self.updater.has_class_mapping(module, name):
                original_module, original_name = module, name
                module, name = self.updater.class_map[(module, name)]
                logger.debug('Modifying state from [%s.%s (v.%s)] to ' + \
                    '[%s.%s (v.%s)]', original_module, original_name, version,
                    module, name, version)
            else:
                break

        # If one exists, call the final class's setstate method. According to
        # standard pickling protocol, this method will apply the state to the
        # instance so our state becomes None so that we don't try to apply our
        # unfinished state to the object.
        fn = getattr(obj, _SETSTATE_NAME, None)
        if fn is not None:
            fn(state)
            result = None
            version = getattr(obj, target_key)
        else:
            result = state

        # Something is wrong if we aren't at our target class and version!
        if module != obj.__class__.__module__ \
            or name != obj.__class__.__name__ \
            or version != target_version:
            from apptools.sweet_pickle import UnpicklingError
            raise UnpicklingError('Unexpected state! Got ' + \
                '[%s.%s (v.%s)] expected [%s.%s (v.%s)]' % (module, name,
                version, obj.__class__.__module__, obj.__class__.__name__,
                target_version))

        return result


    ### protected interface ##################################################

    def _add_unpickler(self, klass, module, name):
        """ Modifies the specified class so that our 'modify_state' method
            is called when its next instance is unpickled.
        """
        logger.debug('Adding unpickler hook to [%s]', klass)

        # Replace the existing setstate method with ours.
        self._backup_setstate(klass)
        m = __replacement_setstate__.__get__(None, klass)
        setattr(klass, _SETSTATE_NAME, m)

        # Add the information necessary to allow this unpickler to run
        setattr(klass, _UNPICKLER_DATA, (self, module, name))


    def _backup_setstate(self, klass):
        """ Backs up the specified class's setstate method.
        """
        # We only need to back it up if it actually exists.
        method = getattr(klass, _SETSTATE_NAME, None)
        if method is not None:
            logger.debug('Backing up method [%s] to [%s] on [%s]',
                _SETSTATE_NAME, _BACKUP_NAME, klass)
            m = method.__get__(None, klass)
            setattr(klass, _BACKUP_NAME, m)


    def _get_target_class(self, module, name):
        """ Returns the class info that the class, within the specified module
            and with the specified name, should be instantiated as according to
            our associated updater.

            This is done in a manner that allows for chaining of class mappings
            but is tolerant of the fact that a mapping away from an
            intermediate class may not be registered until an attempt is made
            to load that class.
        """
        # Keep a record of the original class asked for.
        original_module, original_name = module, name

        # Iterate through any mappings in a manner that allows us to detect any
        # infinite loops.
        visited = []
        while self.updater.has_class_mapping(module, name):
            if (module, name) in visited:
                from apptools.sweet_pickle import UnpicklingError
                raise UnpicklingError('Detected infinite loop in class ' + \
                    'mapping from [%s.%s] to [%s.%s] within Updater [%s]' % \
                    (original_module, original_name, module, name,
                    self.updater))
            visited.append( (module, name) )

            # Get the mapping for the current class and try loading the class
            # to ensure any mappings away from it are registered.
            module, name = self.updater.class_map[(module, name)]
            try:
                super(VersionedUnpickler, self).find_class(module, name)
            except:
                logger.exception("_get_target_class can't find: %s" % (module, name))
                pass

        return module, name


    def _has_state_function(self, module, name):
        """ Returns True if the updater contains any state functions that could
            be called by unpickling an instance of the class identified by the
            specified module and name.

            Note: If we had a version number we could tell for sure, but we
            don't have one so we'll have to settle for 'could' be called.
        """
        result = False

        # Iterate through all the class mappings the requested class would
        # go through.  If any of them have a state function, then we've
        # determined our answer and can stop searching.
        #
        # Note we don't need to check for infinite loops because we're only
        # ever called after '_get_target_class' which detects the infinite
        # loops.
        while not result:
            result = self.updater.has_state_function(module, name)
            if not result:
                if self.updater.has_class_mapping(module, name):
                    module, name = self.updater.class_map[(module, name)]
                else:
                    break

        return result


    def _remove_unpickler(self, klass):
        """ Restores the specified class to its unmodified state.  Meaning
            we won't get called when its next instance is unpickled.
        """
        logger.debug('Removing unpickler hook from [%s]', klass)

        # Restore the backed up setstate method
        self._restore_setstate(klass)

        # Remove the unpickling data attached to the class.  This ensures we
        # don't pollute the 'real' attributes of the class.
        delattr(klass, _UNPICKLER_DATA)


    def _restore_setstate(self, klass):
        """ Restores the original setstate method back to its rightful place.
        """
        # We only need to restore if the backup actually exists.
        method = getattr(klass, _BACKUP_NAME, None)
        if method is not None:
            logger.debug('Restoring method [%s] to [%s] on [%s]',
                _BACKUP_NAME, _SETSTATE_NAME, klass)
            delattr(klass, _BACKUP_NAME)
            m = method.__get__(None, klass)
            setattr(klass, _SETSTATE_NAME, m)

        # Otherwise, we simply remove our setstate.
        else:
            delattr(klass, _SETSTATE_NAME)


### EOF ######################################################################




#------------------------------------------------------------------------------
# Copyright (c) 2013, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
#------------------------------------------------------------------------------
""" Base classes for adapters.

Adapters do not have to inherit from these classes, as long as their
constructor takes the object to be adapted as the first and only
*positional* argument.

"""


from traits.has_traits import HasTraits
from traits.trait_types import Any
from traits.util.deprecated import deprecated


class PurePythonAdapter(object):
    """ Base class for pure Python adapters. """

    def __init__(self, adaptee):
        """ Constructor. """

        self.adaptee = adaptee

        return


class Adapter(HasTraits):
    """ Base class for adapters with traits. """

    def __init__(self, adaptee, **traits):
        """ Constructor. """

        traits['adaptee'] = adaptee
        super(Adapter, self).__init__(**traits)

        return

    adaptee = Any


def adapts(from_, to, extra=None, factory=None, cached=False, when=''):
    """ A class advisor for declaring adapters.

    Parameters
    ----------
    from_ : type or interface
        What the adapter adapts *from*, or a list of such types or interfaces
        (the '_' suffix is used because 'from' is a Python keyword).
    to : type or interface
        What the adapter adapts *to*, or a list of such types or interfaces.
    factory : callable
        An (optional) factory for actually creating the adapters. This is
        any callable that takes a single argument which is the object to
        be adapted. The factory should return an adapter if it can
        perform the adaptation and **None** if it cannot.
    cached : bool
        Should the adapters be cached? If an adapter is cached, then the
        factory will produce at most one adapter per instance.
    when : str
        A Python expression that selects which instances of a particular type
        can be adapted by this factory. The expression is evaluated in a
        namespace that contains a single name *adaptee*, which is bound to the
        object to be adapted (e.g., 'adaptee.is_folder').

    Note
    ----
    The ``cached`` and ``when`` arguments are ignored if ``factory`` is
    specified.

    """

    from traits.adaptation.api import register_factory
    from traits.adaptation.cached_adapter_factory import CachedAdapterFactory
    from traits.protocols.advice import addClassAdvisor

    if extra is not None:
        adapter, from_, to = from_, to, extra
    else:
        adapter = None

    @deprecated("use the 'register_factory' function from 'traits.api' instead")
    def callback(klass):
        """ Called when the class has been created. """

        # At this point:-
        #
        # klass is the callable (usually a class) that takes one argument (the
        # adaptee) and returns an appropriate adapter (or None if the adaptation
        # is not possible).

        # What the adapters created by the factory will adapt from.
        if type(from_) is not list:
            from_protocols = [from_]

        else:
            from_protocols = from_

        # What the adapters created by the factory will adapt to.
        if type(to) is not list:
            to_protocols = [to]

        else:
            to_protocols = to

        if factory is None:
            # If the adapter is cached or has a 'when' expression then create a
            # default factory:

            adapter_factory = klass

            if when != '':
                def _conditional_factory(adaptee, *args, **kw):
                    namespace = {'adaptee': adaptee}

                    if eval(when, namespace, namespace):
                        return klass(adaptee, *args, **kw)

                    return None

                adapter_factory = _conditional_factory

            if cached:
                adapter_factory = CachedAdapterFactory(factory=adapter_factory)

        else:
            adapter_factory = factory

        for from_protocol in from_protocols:
            for to_protocol in to_protocols:
                register_factory(adapter_factory, from_protocol, to_protocol)

        for to_protocol in to_protocols:
            # We cannot register adapter factories that are functions. (This is
            # ony relevant when using 'adapts' as a function.
            if isinstance(klass, type):
                # We use type(to_protocol) in case the to_protocols implements
                # its own 'register' method which overrides the ABC method.
                type(to_protocol).register(to_protocol, klass)

        return klass

    if adapter is not None:
        callback(adapter)

    else:
        addClassAdvisor(callback)

#### EOF ######################################################################



# Standard library imports.
import os

# Enthought library imports.
from envisage.api import ExtensionPoint, Plugin
from envisage.ui.tasks.api import TaskFactory
from traits.api import List, TraitError
from envisage.ui.tasks.tasks_plugin import TasksPlugin
from traits.etsconfig.api import ETSConfig
from apptools.preferences.api import Preferences, PreferencesHelper
from pyface.tasks.action.api import SMenuBar, SMenu, SchemaAddition


class OmnivoreTasksPlugin(TasksPlugin):
    # Override the default task extensions that supply redundant Exit and
    # Preferences menu items and the default dock pane viewing group.
    def _my_task_extensions_default(self):
        return []


class FrameworkPlugin(Plugin):
    """ The sample framework plugin.
    """

    #### convenience functions

    def task_factories_from_tasks(self, tasks):
        # Create task factories for each task such that each factory will have
        # the same id as the task
        factories = []
        for cls in tasks:
            task = cls()
            factory = TaskFactory(id=task.id, name=task.name, factory=cls)
            factories.append(factory)

        return factories

    def get_helper(self, helper_object, debug=True):
        return self.application.get_preferences(helper_object, debug)

    def set_plugin_data(self, data):
        """Store some plugin data in the application so that objects outside
        the plugin can have access to it
        
        The data is stored in a dict keyed on the plugin's id, so make
        sure plugins don't have the same id.  (Perhaps this is enforced in
        Enthought, haven't checked yet.)
        """
        self.application.plugin_data[self.id] = data

    def get_plugin_data(self):
        """Return the plugin data previously stored by a call to
        :py:meth:`set_plugin_data`
        
        """
        return self.application.plugin_data[self.id]

    def fire_plugin_event(self, data=None):
        """Send a plugin event.
        
        Plugin events will get fired to all who listen for the
        'application:plugin_event' event, e.g.:
        
            @on_trait_change('application:plugin_event')
        
        The event handler will be passed a tuple of the plugin's ID and some
        data.  All handlers that listen for this event will get called, so
        check the first item in the tuple and for the desired plugin ID.
        """
        self.application.plugin_event = (self.id, data)


class OmnivoreMainPlugin(FrameworkPlugin):
    """ The sample framework plugin.
    """

    # Extension point IDs.
    PREFERENCES       = 'envisage.preferences'
    PREFERENCES_PANES = 'envisage.ui.tasks.preferences_panes'
    TASKS             = 'envisage.ui.tasks.tasks'
    OSX_MINIMAL_MENU = 'omnivore.osx_minimal_menu'

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'omnivore.framework.plugin'

    # The plugin's name (suitable for displaying to the user).
    name = 'Omnivore'

    #### Contributions to extension points made by this plugin ################

    preferences = List(contributes_to=PREFERENCES)
    preferences_panes = List(contributes_to=PREFERENCES_PANES)
    tasks = List(contributes_to=TASKS)
    osx_actions = List(contributes_to=OSX_MINIMAL_MENU)

    ###########################################################################
    # Protected interface.
    ###########################################################################

    def _preferences_default(self):
        filename = os.path.join(ETSConfig.application_home, 'preferences.ini')
        if not os.path.exists(filename):
            fh = open(filename, "wb")
            fh.close()
        return [ 'file://' + filename ]

    def _preferences_panes_default(self):
        from preferences import FrameworkPreferencesPane
        from omnivore.tasks.text_edit import TextEditPreferencesPane
        from omnivore.tasks.image_edit import ImageEditPreferencesPane
        from omnivore.tasks.html_view import HtmlViewPreferencesPane
        return [ FrameworkPreferencesPane, TextEditPreferencesPane, ImageEditPreferencesPane, HtmlViewPreferencesPane]

    def _tasks_default(self):
        from omnivore.tasks.text_edit import TextEditTask
        from omnivore.tasks.image_edit import ImageEditTask
        from omnivore.tasks.html_view import HtmlViewTask

        return self.task_factories_from_tasks([
            TextEditTask,
            ImageEditTask,
            HtmlViewTask,
            ])

    def _osx_actions_default(self):
        from omnivore.framework.actions import NewFileGroup

        submenu = lambda: SMenu(
            id='NewFileSubmenu', name="New"
        )
        actions = [
            SchemaAddition(factory=submenu,
                           path='MenuBar/File',
                           before="OpenGroup"
                           ),
            SchemaAddition(factory=NewFileGroup,
                           path='MenuBar/File/NewFileSubmenu',
                           ),
            ]

        return actions



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

""" Tests the class mapping functionality of the enthought.pickle
    framework.
"""

# Standard library imports.
import unittest

# Enthought library imports
import apptools.sweet_pickle as sweet_pickle
from apptools.sweet_pickle.global_registry import _clear_global_registry


##############################################################################
# Classes to use within the tests
##############################################################################

# Need complete package name so that mapping matches correctly.
# The problem here is the Python loader that will load the same module with
# multiple names in sys.modules due to relative naming. Nice.
from apptools.sweet_pickle.tests.class_mapping_classes import Foo, Bar, Baz

##############################################################################
# class 'ClassMappingTestCase'
##############################################################################

class ClassMappingTestCase(unittest.TestCase):
    """ Tests the class mapping functionality of the apptools.sweet_pickle
        framework.
    """

    ##########################################################################
    # 'TestCase' interface
    ##########################################################################

    ### public interface #####################################################

    def setUp(self):
        """ Creates the test fixture.

            Overridden here to ensure each test starts with an empty global
            registry.
        """
        # Clear the global registry
        _clear_global_registry()

        # Cache a reference to the new global registry
        self.registry = sweet_pickle.get_global_registry()


    ##########################################################################
    # 'ClassMappingTestCase' interface
    ##########################################################################

    ### public interface #####################################################

    def test_infinite_loop_detection(self):
        """ Validates that the class mapping framework detects infinite
            loops of class mappings.
        """
        # Add mappings to the registry
        self.registry.add_mapping_to_class(Foo.__module__, Foo.__name__, Bar)
        self.registry.add_mapping_to_class(Bar.__module__, Bar.__name__, Baz)
        self.registry.add_mapping_to_class(Baz.__module__, Baz.__name__, Foo)

        # Validate that an exception is raised when trying to unpickle an
        # instance anywhere within the circular definition.
        def fn(o):
            sweet_pickle.loads(sweet_pickle.dumps(o))
        self.assertRaises(sweet_pickle.UnpicklingError, fn, Foo())
        self.assertRaises(sweet_pickle.UnpicklingError, fn, Bar())
        self.assertRaises(sweet_pickle.UnpicklingError, fn, Baz())


    def test_unpickled_class_mapping(self):

        # Add the mappings to the registry
        self.registry.add_mapping_to_class(Foo.__module__, Foo.__name__, Bar)
        self.registry.add_mapping_to_class(Bar.__module__, Bar.__name__, Baz)

        # Validate that unpickling the first class gives us an instance of
        # the third class.
        start = Foo()
        end = sweet_pickle.loads(sweet_pickle.dumps(start))
        self.assertEqual(True, isinstance(end, Baz))

        # Validate that unpickling the second class gives us an instance of
        # the third class.
        start = Bar()
        end = sweet_pickle.loads(sweet_pickle.dumps(start))
        self.assertEqual(True, isinstance(end, Baz))


if __name__ == "__main__":
    unittest.main()


### EOF ######################################################################



"""
fs.expose.django
================

Use an FS object for Django File Storage

This module exposes the class "FSStorage", a simple adapter for using FS
objects as Django storage objects.  Simply include the following lines
in your settings.py::

    DEFAULT_FILE_STORAGE = fs.expose.django_storage.FSStorage
    DEFAULT_FILE_STORAGE_FS = OSFS('foo/bar')  # Or whatever FS


"""

from django.conf import settings
from django.core.files.storage import Storage
from django.core.files import File

from fs.path import abspath, dirname
from fs.errors import convert_fs_errors, ResourceNotFoundError

class FSStorage(Storage):
    """Expose an FS object as a Django File Storage object."""

    def __init__(self, fs=None, base_url=None):
        """
        :param fs: an FS object
        :param base_url: The url to prepend to the path
         
        """
        if fs is None:
            fs = settings.DEFAULT_FILE_STORAGE_FS
        if base_url is None:
            base_url = settings.MEDIA_URL
        base_url = base_url.rstrip('/')
        self.fs = fs
        self.base_url = base_url

    def exists(self, name):
        return self.fs.isfile(name)

    def path(self, name):
        path = self.fs.getsyspath(name)
        if path is None:
            raise NotImplementedError
        return path

    @convert_fs_errors
    def size(self, name):
        return self.fs.getsize(name)

    @convert_fs_errors
    def url(self, name):
        return self.base_url + abspath(name)

    @convert_fs_errors
    def _open(self, name, mode):
        return File(self.fs.open(name, mode))

    @convert_fs_errors
    def _save(self, name, content):
        self.fs.makedir(dirname(name), allow_recreate=True, recursive=True)
        self.fs.setcontents(name, content)
        return name

    @convert_fs_errors
    def delete(self, name):
        try:
            self.fs.remove(name)
        except ResourceNotFoundError:
            pass





# Standard library imports.
import os.path

# Enthought library imports.
from envisage.api import ExtensionPoint, Plugin, ServiceOffer
from traits.api import List, Instance

from i_file_recognizer import IFileRecognizer, IFileRecognizerDriver

import logging
log = logging.getLogger(__name__)


class FileTypePlugin(Plugin):
    """ Plugin for identifying file types
    """

    # The Ids of the extension points that this plugin offers.
    RECOGNIZER = 'omnivore.file_recognizer'

    # Extension point IDs.
    SERVICE_OFFERS    = 'envisage.service_offers'

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'omnivore.file_type.plugin'

    # The plugin's name (suitable for displaying to the user).
    name = 'File Type'

    #### Extension points offered by this plugin ##############################

    recognizers = ExtensionPoint(
        List(Instance(IFileRecognizer)), id=RECOGNIZER, desc="""
    
    This extension point allows you to contribute file scanners that determine
    MIME types from a byte stream.
    
        """
    )

    #### Contributions to extension points made by this plugin ################

    service_offers = List(contributes_to=SERVICE_OFFERS)

    ###########################################################################
    # Protected interface.
    ###########################################################################

    def _service_offers_default(self):
        """ Trait initializer. """

        log.debug("in _service_offers_default")
        offer1 = ServiceOffer(
            protocol = 'omnivore.file_type.i_file_recognizer.IFileRecognizerDriver',
            factory  = self._create_file_recognizer_driver_service
        )

        return [offer1]

    def _create_file_recognizer_driver_service(self):
        """ Factory method for the File Recognizer Driver service. """

        log.debug("known recognizers: %s" % str(self.recognizers))

        # Lazy importing, even though this is a fundamental service and
        # therefore doesn't buy us anything.  But as an example it's useful.
        from .driver import FileRecognizerDriver
        return FileRecognizerDriver(recognizers=self.recognizers, application=self.application)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
from context_monitor import ContextMonitor
from context_node_type import ContextNodeType
from explorer import Explorer, explore
from naming_node_manager import NamingNodeManager
from naming_tree import NamingTree
from naming_tree_model import NamingTreeModel
from object_node_type import ObjectNodeType



#-------------------------------------------------------------------------------
#
#  Defines the ITemplateDataContext interface for accessing a named collection
#  of data that can be bound to a templatized object when converting it to a
#  'live' set of objects.
#
#  Written by: David C. Morrill
#
#  Date: 07/26/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines the ITemplateDataContext interface for accessing a named collection
    of data that can be bound to a templatized object when converting it to a
    'live' set of objects.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api\
    import Interface, List, Str

#-------------------------------------------------------------------------------
#  'ITemplateDataContext' interface:
#-------------------------------------------------------------------------------

class ITemplateDataContext ( Interface ):
    """ Defines the ITemplateDataContext interface for accessing a named
        collection of data that can be bound to a templatized object when
        converting it to a 'live' set of objects.
    """

    # The path to this data context (does not include the 'data_context_name'):
    data_context_path = Str

    # The name of the data context:
    data_context_name = Str

    # A list of the names of the data values in this context:
    data_context_values = List( Str )

    # The list of the names of the sub-contexts of this context:
    data_contexts = List( Str )

    def get_data_context_value ( self, name ):
        """ Returns the data value with the specified *name*. Raises a
            **ITemplateDataContextError** if *name* is not defined as a data
            value in the context.

            Parameters
            ----------
            name : A string specifying the name of the context data value to
                be returned.

            Returns
            -------
            The data value associated with *name* in the context. The type of
            the data is application dependent.

            Raises **ITemplateDataContextError** if *name* is not associated
            with a data value in the context.
        """

    def get_data_context ( self, name ):
        """ Returns the **ITemplateDataContext** value associated with the
            specified *name*. Raises **ITemplateDataContextError** if *name* is
            not defined as a data context in the context.

            Parameters
            ----------
            name : A string specifying the name of the data context to be
                returned.

            Returns
            -------
            The **ITemplateDataContext** associated with *name* in the context.

            Raises **ITemplateDataContextError** if *name* is not associated
            with a data context in the context.
        """

#-------------------------------------------------------------------------------
#  'ITemplateDataContextError' exception:
#-------------------------------------------------------------------------------

class ITemplateDataContextError ( Exception ):
    """ The exception class associated with the **ITemplateDataContext**
        interface.
    """




#!/usr/bin/env python

"""
Tests for adding a data attribute to a UGrid object.

"""

from __future__ import (absolute_import, division, print_function)

import numpy as np
import pytest

from pyugrid import UVar

from .utilities import two_triangles


def test_add_all_data():
    """You should not be able add a data dict directly."""
    grid = two_triangles()

    assert grid.data == {}

    with pytest.raises(AttributeError):
        grid.data = {'depth': UVar('depth',
                                   location='node',
                                   data=[1.0, 2.0, 3.0, 4.0])}


def test_add_node_data():
    grid = two_triangles()

    # Create a UVar object for the depths:
    depths = UVar('depth', location='node', data=[1.0, 2.0, 3.0, 4.0])
    depths.attributes['units'] = 'm'
    depths.attributes['standard_name'] = 'sea_floor_depth'
    depths.attributes['positive'] = 'down'

    grid.add_data(depths)

    assert grid.data['depth'].name == 'depth'
    assert grid.data['depth'].attributes['units'] == 'm'
    assert np.array_equal(grid.data['depth'].data, [1.0, 2.0, 3.0, 4.0])


def test_add_node_data_wrong():
    """Too short an array."""

    grid = two_triangles()

    # Create a UVar object for the depths:
    depths = UVar('depth', location='node', data=[1.0, 2.0, 3.0])

    with pytest.raises(ValueError):
        grid.add_data(depths)


def test_add_face_data():
    grid = two_triangles()

    # Create a UVar object for velocity:
    u_vel = UVar('u', location='face', data=[1.0, 2.0])
    u_vel.attributes['units'] = 'm/s'
    u_vel.attributes['standard_name'] = 'eastward_sea_water_velocity'

    grid.add_data(u_vel)

    assert grid.data['u'].name == 'u'
    assert grid.data['u'].attributes['units'] == 'm/s'
    assert np.array_equal(grid.data['u'].data, [1.0, 2.0])


def test_add_face_data_wrong():
    """Too short an array."""

    grid = two_triangles()

    # Create a UVar object for velocity:
    u_vel = UVar('u', location='face', data=[1.0])

    with pytest.raises(ValueError):
        grid.add_data(u_vel)


def test_add_edge_data():
    grid = two_triangles()

    # Create a UVar object for velocity:
    bnds = UVar('bounds', location='edge', data=[0, 1, 0, 0, 1])
    bnds.attributes['standard_name'] = 'boundary type'

    grid.add_data(bnds)

    assert grid.data['bounds'].name == 'bounds'
    assert np.array_equal(grid.data['bounds'].data, [0, 1, 0, 0, 1])


def test_add_edge_data_wrong():
    """Too long an array."""
    grid = two_triangles()

    # Create a UVar object for velocity:
    # a miss-matched set.
    bnds = UVar('bounds', location='edge', data=[0, 1, 0, 0, 1, 3, 3])
    with pytest.raises(ValueError):
        grid.add_data(bnds)


def test_add_boundary_data():
    grid = two_triangles()

    # Add the boundary definitions:
    grid.boundaries = [(0, 1),
                       (0, 2),
                       (1, 3),
                       (2, 3)]

    # Create a UVar object for boundary conditions:
    bnds = UVar('bounds', location='boundary', data=[0, 1, 0, 0, 1])
    bnds.attributes['long_name'] = 'model boundary conditions'

    # Wrong size for data.
    with pytest.raises(ValueError):
        grid.add_data(bnds)

    # Correct data.
    bnds.data = [0, 1, 0, 0]
    grid.add_data(bnds)

    assert grid.data['bounds'].name == 'bounds'
    assert np.array_equal(grid.data['bounds'].data, [0, 1, 0, 0])



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/14/2007
#
#-------------------------------------------------------------------------

""" Defines the default DockWindow theme.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from .dock_window_theme import DockWindowTheme

from .theme import Theme

#-------------------------------------------------------------------------
#  Define the default theme:
#-------------------------------------------------------------------------

# The original DockWindows UI redone as a theme:
default_dock_window_theme = DockWindowTheme(
    use_theme_color=False,
    tab_active=Theme('@std:tab_active',
                     label=(0, -3), content=(7, 6, 0, 0)),
    tab_inactive=Theme('@std:tab_inactive',
                       label=(0, -1), content=(5, 0)),
    tab_hover=Theme('@std:tab_hover',
                    label=(0, -2), content=(5, 0)),
    tab_background=Theme('@std:tab_background'),
    tab=Theme('@std:tab',
              content=0, label=(-7, 0)),
    vertical_splitter=Theme('@std:vertical_splitter',
                            content=0, label=(0, -25)),
    horizontal_splitter=Theme('@std:horizontal_splitter',
                              content=0, label=(-24, 0)),
    vertical_drag=Theme('@std:vertical_drag',
                        content=(0, 10)),
    horizontal_drag=Theme('@std:horizontal_drag',
                          content=(10, 0))
)



import unittest

try:
  from pyface.ui.wx.grid.api \
       import GridRow, GridColumn, SimpleGridModel
except ImportError:
    wx_available = False
else:
    wx_available = True


@unittest.skipUnless(wx_available, "Wx is not available")
class CompositeGridModelTestCase( unittest.TestCase ):

    def setUp(self):

        self.model = SimpleGridModel(data=[[None,2],[3,4]],
                                     rows=[GridRow(label='foo'),
                                           GridRow(label='bar')],
                                     columns=[GridColumn(label='cfoo'),
                                              GridColumn(label='cbar')]
                                     )

        return

    def test_get_column_count(self):

        self.assertEqual(self.model.get_column_count(), 2)

        return

    def test_get_row_count(self):

        self.assertEqual(self.model.get_row_count(), 2)

        return

    def test_get_row_name(self):

        # Regardless of the rows specified in the composed models, the
        # composite model returns its own rows.
        self.assertEqual(self.model.get_row_name(0), 'foo')
        self.assertEqual(self.model.get_row_name(1), 'bar')

        return

    def test_get_column_name(self):

        self.assertEqual(self.model.get_column_name(0), 'cfoo')
        self.assertEqual(self.model.get_column_name(1), 'cbar')

        return

    def test_get_value(self):

        self.assertEqual(self.model.get_value(0,0), None)
        self.assertEqual(self.model.get_value(0,1), 2)
        self.assertEqual(self.model.get_value(1,0), 3)
        self.assertEqual(self.model.get_value(1,1), 4)

        return

    def test_is_cell_empty(self):

        rows = self.model.get_row_count()
        columns = self.model.get_column_count()

        self.assertEqual(self.model.is_cell_empty(0,0), True,
                         "Cell containing None should be empty.")
        self.assertEqual(self.model.is_cell_empty(10,10), True,
                         "Cell outside the range of values should be empty.")

        return


#### EOF ######################################################################



"""

  fs.tests.test_wrapfs:  testcases for FS wrapper implementations

"""

import unittest
from fs.tests import FSTestCases, ThreadingTestCases

import os
import sys
import shutil
import tempfile

from fs import osfs
from fs.errors import * 
from fs.path import *
from fs.utils import remove_all
from fs import wrapfs

import six
from six import PY3, b

class TestWrapFS(unittest.TestCase, FSTestCases, ThreadingTestCases):
    
    #__test__ = False
    
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp(u"fstest")
        self.fs = wrapfs.WrapFS(osfs.OSFS(self.temp_dir))

    def tearDown(self):
        shutil.rmtree(self.temp_dir)
        self.fs.close()

    def check(self, p):
        return os.path.exists(os.path.join(self.temp_dir, relpath(p)))


from fs.wrapfs.lazyfs import LazyFS
class TestLazyFS(unittest.TestCase, FSTestCases, ThreadingTestCases):
    
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp(u"fstest")
        self.fs = LazyFS((osfs.OSFS,(self.temp_dir,)))

    def tearDown(self):
        shutil.rmtree(self.temp_dir)
        self.fs.close()

    def check(self, p):
        return os.path.exists(os.path.join(self.temp_dir, relpath(p)))


from fs.wrapfs.limitsizefs import LimitSizeFS
class TestLimitSizeFS(TestWrapFS):

    _dont_retest = TestWrapFS._dont_retest + ("test_big_file",)
    
    def setUp(self):
        super(TestLimitSizeFS,self).setUp()
        self.fs = LimitSizeFS(self.fs,1024*1024*2)  # 2MB limit

    def tearDown(self):
        remove_all(self.fs, "/")
        self.assertEquals(self.fs.cur_size,0)
        super(TestLimitSizeFS,self).tearDown()
        self.fs.close()

    def test_storage_error(self):
        total_written = 0
        for i in xrange(1024*2):
            try:
                total_written += 1030
                self.fs.setcontents("file %i" % i, b("C")*1030)
            except StorageSpaceError:
                self.assertTrue(total_written > 1024*1024*2)
                self.assertTrue(total_written < 1024*1024*2 + 1030)
                break
        else:
            self.assertTrue(False,"StorageSpaceError not raised")           


from fs.wrapfs.hidedotfilesfs import HideDotFilesFS
class TestHideDotFilesFS(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp(u"fstest")
        open(os.path.join(self.temp_dir, u".dotfile"), 'w').close()
        open(os.path.join(self.temp_dir, u"regularfile"), 'w').close()
        os.mkdir(os.path.join(self.temp_dir, u".dotdir"))
        os.mkdir(os.path.join(self.temp_dir, u"regulardir"))
        self.fs = HideDotFilesFS(osfs.OSFS(self.temp_dir))

    def tearDown(self):
        shutil.rmtree(self.temp_dir)
        self.fs.close()

    def test_hidden(self):
        self.assertEquals(len(self.fs.listdir(hidden=False)), 2)
        self.assertEquals(len(list(self.fs.ilistdir(hidden=False))), 2)

    def test_nonhidden(self):
        self.assertEquals(len(self.fs.listdir(hidden=True)), 4)
        self.assertEquals(len(list(self.fs.ilistdir(hidden=True))), 4)

    def test_default(self):
        self.assertEquals(len(self.fs.listdir()), 2)
        self.assertEquals(len(list(self.fs.ilistdir())), 2)





#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#------------------------------------------------------------------------------
""" Initialize this backend.
"""

from pyface.base_toolkit import Toolkit

toolkit_object = Toolkit('null', 'pyface.ui.null')



"""

  fs.tests.test_xattr:  testcases for extended attribute support

"""

import unittest
import os

from fs.path import *
from fs.errors import *
from fs.tests import FSTestCases

from six import b 

class XAttrTestCases:
    """Testcases for filesystems providing extended attribute support.

    This class should be used as a mixin to the unittest.TestCase class
    for filesystems that provide extended attribute support.
    """

    def test_getsetdel(self):
        def do_getsetdel(p):
            self.assertEqual(self.fs.getxattr(p,"xattr1"),None)
            self.fs.setxattr(p,"xattr1","value1")
            self.assertEqual(self.fs.getxattr(p,"xattr1"),"value1")
            self.fs.delxattr(p,"xattr1")
            self.assertEqual(self.fs.getxattr(p,"xattr1"),None)
        self.fs.setcontents("test.txt",b("hello"))
        do_getsetdel("test.txt")
        self.assertRaises(ResourceNotFoundError,self.fs.getxattr,"test2.txt","xattr1")
        self.fs.makedir("mystuff")
        self.fs.setcontents("/mystuff/test.txt",b(""))
        do_getsetdel("mystuff")
        do_getsetdel("mystuff/test.txt")

    def test_list_xattrs(self):
        def do_list(p):
            self.assertEquals(sorted(self.fs.listxattrs(p)),[])
            self.fs.setxattr(p,"xattr1","value1")
            self.assertEquals(self.fs.getxattr(p,"xattr1"),"value1")
            self.assertEquals(sorted(self.fs.listxattrs(p)),["xattr1"])
            self.assertTrue(isinstance(self.fs.listxattrs(p)[0],unicode))
            self.fs.setxattr(p,"attr2","value2")
            self.assertEquals(sorted(self.fs.listxattrs(p)),["attr2","xattr1"])
            self.assertTrue(isinstance(self.fs.listxattrs(p)[0],unicode))
            self.assertTrue(isinstance(self.fs.listxattrs(p)[1],unicode))
            self.fs.delxattr(p,"xattr1")
            self.assertEquals(sorted(self.fs.listxattrs(p)),["attr2"])
            self.fs.delxattr(p,"attr2")
            self.assertEquals(sorted(self.fs.listxattrs(p)),[])
        self.fs.setcontents("test.txt",b("hello"))
        do_list("test.txt")
        self.fs.makedir("mystuff")
        self.fs.setcontents("/mystuff/test.txt",b(""))
        do_list("mystuff")
        do_list("mystuff/test.txt")

    def test_copy_xattrs(self):
        self.fs.setcontents("a.txt",b("content"))
        self.fs.setxattr("a.txt","myattr","myvalue")
        self.fs.setxattr("a.txt","testattr","testvalue")
        self.fs.makedir("stuff")
        self.fs.copy("a.txt","stuff/a.txt")
        self.assertTrue(self.fs.exists("stuff/a.txt"))
        self.assertEquals(self.fs.getxattr("stuff/a.txt","myattr"),"myvalue")
        self.assertEquals(self.fs.getxattr("stuff/a.txt","testattr"),"testvalue")
        self.assertEquals(self.fs.getxattr("a.txt","myattr"),"myvalue")
        self.assertEquals(self.fs.getxattr("a.txt","testattr"),"testvalue")
        self.fs.setxattr("stuff","dirattr","a directory")
        self.fs.copydir("stuff","stuff2")
        self.assertEquals(self.fs.getxattr("stuff2/a.txt","myattr"),"myvalue")
        self.assertEquals(self.fs.getxattr("stuff2/a.txt","testattr"),"testvalue")
        self.assertEquals(self.fs.getxattr("stuff2","dirattr"),"a directory")
        self.assertEquals(self.fs.getxattr("stuff","dirattr"),"a directory")

    def test_move_xattrs(self):
        self.fs.setcontents("a.txt",b("content"))
        self.fs.setxattr("a.txt","myattr","myvalue")
        self.fs.setxattr("a.txt","testattr","testvalue")
        self.fs.makedir("stuff")
        self.fs.move("a.txt","stuff/a.txt")
        self.assertTrue(self.fs.exists("stuff/a.txt"))
        self.assertEquals(self.fs.getxattr("stuff/a.txt","myattr"),"myvalue")
        self.assertEquals(self.fs.getxattr("stuff/a.txt","testattr"),"testvalue")
        self.fs.setxattr("stuff","dirattr","a directory")
        self.fs.movedir("stuff","stuff2")
        self.assertEquals(self.fs.getxattr("stuff2/a.txt","myattr"),"myvalue")
        self.assertEquals(self.fs.getxattr("stuff2/a.txt","testattr"),"testvalue")
        self.assertEquals(self.fs.getxattr("stuff2","dirattr"),"a directory")

    def test_remove_file(self):
        def listxattrs(path):
            return list(self.fs.listxattrs(path))
        #  Check that xattrs aren't preserved after a file is removed
        self.fs.createfile("myfile")
        self.assertEquals(listxattrs("myfile"),[])
        self.fs.setxattr("myfile","testattr","testvalue")
        self.assertEquals(listxattrs("myfile"),["testattr"])
        self.fs.remove("myfile")
        self.assertRaises(ResourceNotFoundError,listxattrs,"myfile")
        self.fs.createfile("myfile")
        self.assertEquals(listxattrs("myfile"),[])
        self.fs.setxattr("myfile","testattr2","testvalue2")
        self.assertEquals(listxattrs("myfile"),["testattr2"])
        self.assertEquals(self.fs.getxattr("myfile","testattr2"),"testvalue2")
        #  Check that removing a file without xattrs still works
        self.fs.createfile("myfile2")
        self.fs.remove("myfile2")
 
    def test_remove_dir(self):
        def listxattrs(path):
            return list(self.fs.listxattrs(path))
        #  Check that xattrs aren't preserved after a dir is removed
        self.fs.makedir("mydir")
        self.assertEquals(listxattrs("mydir"),[])
        self.fs.setxattr("mydir","testattr","testvalue")
        self.assertEquals(listxattrs("mydir"),["testattr"])
        self.fs.removedir("mydir")
        self.assertRaises(ResourceNotFoundError,listxattrs,"mydir")
        self.fs.makedir("mydir")
        self.assertEquals(listxattrs("mydir"),[])
        self.fs.setxattr("mydir","testattr2","testvalue2")
        self.assertEquals(listxattrs("mydir"),["testattr2"])
        self.assertEquals(self.fs.getxattr("mydir","testattr2"),"testvalue2")
        #  Check that removing a dir without xattrs still works
        self.fs.makedir("mydir2")
        self.fs.removedir("mydir2")
        #  Check that forcibly removing a dir with xattrs still works
        self.fs.makedir("mydir3")
        self.fs.createfile("mydir3/testfile")
        self.fs.removedir("mydir3",force=True)
        self.assertFalse(self.fs.exists("mydir3"))


from fs.xattrs import ensure_xattrs

from fs import tempfs
class TestXAttr_TempFS(unittest.TestCase,FSTestCases,XAttrTestCases):

    def setUp(self):
        fs = tempfs.TempFS()
        self.fs = ensure_xattrs(fs)

    def tearDown(self):
        try:
            td = self.fs._temp_dir
        except AttributeError:
            td = self.fs.wrapped_fs._temp_dir
        self.fs.close()
        self.assert_(not os.path.exists(td))

    def check(self, p):
        try:
            td = self.fs._temp_dir
        except AttributeError:
            td = self.fs.wrapped_fs._temp_dir
        return os.path.exists(os.path.join(td, relpath(p)))


from fs import memoryfs
class TestXAttr_MemoryFS(unittest.TestCase,FSTestCases,XAttrTestCases):

    def setUp(self):
        self.fs = ensure_xattrs(memoryfs.MemoryFS())

    def check(self, p):
        return self.fs.exists(p)





#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/21/2002
#
#  Refactored into a separate module: 07/04/2003
#
#------------------------------------------------------------------------------

""" Defines common, low-level capabilities needed by the Traits package.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from __future__ import absolute_import

import os
import sys
from os import getcwd
from os.path import dirname, exists, join

from . import _py2to3

# backwards compatibility: trait_base used to provide a patched enumerate
enumerate = enumerate

# Set the Python version being used:
vi = sys.version_info
python_version = vi[0] + (float( vi[1] ) / 10.0)

try:
    from traits.etsconfig.api import ETSConfig
except:
    # If the ETSConfig package is not available, fake it:
    class ETSConfig ( object ):

        #-----------------------------------------------------------------------
        #  'object' interface:
        #-----------------------------------------------------------------------

        def __init__ ( self ):
            """ Constructor.

                Note that this constructor can only ever be called from within
                this module, since we don't expose the class.
            """
            # Shadow attributes for properties:
            self._application_data = None
            self._toolkit          = None

            return

        #-----------------------------------------------------------------------
        #  'ETSConfig' interface:
        #-----------------------------------------------------------------------

        #-- Property Implementations -------------------------------------------

        def _get_application_data ( self ):
            """ Property getter.

                This is a directory that applications and packages can safely
                write non-user accessible data to i.e. configuration
                information, preferences etc.

                Do not put anything in here that the user might want to navigate
                to (e.g. projects, user data files, etc).

                The actual location differs between operating systems.
            """
            if self._application_data is None:
                self._application_data = self._initialize_application_data()

            return self._application_data

        def _set_application_data ( self, application_data ):
            """ Property setter.
            """
            self._application_data = application_data

        application_data = property( _get_application_data,
                                     _set_application_data )

        def _get_toolkit ( self ):
            """
            Property getter for the GUI toolkit.  The value returned is, in
            order of preference: the value set by the application; the value
            passed on the command line using the '-toolkit' option; the value
            specified by the 'ETS_TOOLKIT' environment variable; otherwise the
            empty string.
            """
            if self._toolkit is None:
                self._toolkit = self._initialize_toolkit()

            return self._toolkit

        def _set_toolkit ( self, toolkit ):
            """
            Property setter for the GUI toolkit.  The toolkit can be set more
            than once, but only if it is the same one each time.  An application
            that is written for a particular toolkit can explicitly set it
            before any other module that gets the value is imported.

            """
            if self._toolkit and (self._toolkit != toolkit):
                raise ValueError( 'Cannot set toolkit to %s because it has '
                         'already been set to %s' % ( toolkit, self._toolkit ) )

            self._toolkit = toolkit

            return

        toolkit = property( _get_toolkit, _set_toolkit )

        #-- Private Methods ----------------------------------------------------

        def _initialize_application_data ( self ):
            """ Initializes the (default) application data directory.
            """
            if sys.platform == 'win32':
                environment_variable = 'APPDATA'
                directory_name       = 'Enthought'

            else:
                environment_variable = 'HOME'
                directory_name       = '.enthought'

            # Lookup the environment variable:
            parent_directory = os.environ.get( environment_variable, None )
            if parent_directory is None:
                raise ValueError( 'Environment variable "%s" not set' %
                                  environment_variable )

            application_data = os.path.join( parent_directory, directory_name )

            # If a file already exists with this name then make sure that it is
            # a directory!
            if os.path.exists( application_data ):
                if not os.path.isdir( application_data ):
                    raise ValueError( 'File "%s" already exists' %
                                      application_data )

            # Otherwise, create the directory:
            else:
                os.makedirs( application_data )

            return application_data

        def _initialize_toolkit ( self ):
            """ Initializes the toolkit.
            """
            # We handle the command line option even though it doesn't have the
            # highest precedence because we always want to remove it from the
            # command line:
            if '-toolkit' in sys.argv:
                opt_idx = sys.argv.index( '-toolkit' )

                try:
                    opt_toolkit = sys.argv[ opt_idx + 1 ]
                except IndexError:
                    raise ValueError( 'The -toolkit command line argument must '
                                      'be followed by a toolkit name' )

                # Remove the option:
                del sys.argv[ opt_idx: opt_idx + 1 ]
            else:
                opt_toolkit = None

            if self._toolkit is not None:
                toolkit = self._toolkit
            elif opt_toolkit is not None:
                toolkit = opt_toolkit
            else:
                toolkit = os.environ.get( 'ETS_TOOLKIT', '' )

            return toolkit

    ETSConfig = ETSConfig()

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

ClassTypes    = _py2to3.ClassTypes

SequenceTypes = ( list, tuple )

ComplexTypes  = ( float, int )

TypeTypes     = ( str,  unicode, int, long, float, complex, list, tuple, dict, bool )

TraitNotifier = '__trait_notifier__'

# The standard Traits property cache prefix:
TraitsCache = '_traits_cache_'

#-------------------------------------------------------------------------------
#  Singleton 'Uninitialized' object:
#-------------------------------------------------------------------------------
Uninitialized = None

class _Uninitialized(object):
    """ The singleton value of this class represents the uninitialized state
        of a trait and is specified as the 'old' value in the trait change
        notification that occurs when the value of a trait is read before being
        set.
    """

    def __new__(cls):
        if Uninitialized is not None:
            return Uninitialized
        else:
            self = object.__new__(cls)
            return self

    def __repr__(self):
        return '<uninitialized>'

    def __reduce_ex__(self, protocol):
        return (_Uninitialized, ())

#: When the first reference to a trait is a 'get' reference, the default value of
#: the trait is implicitly assigned and returned as the value of the trait.
#: Because of this implicit assignment, a trait change notification is
#: generated with the Uninitialized object as the 'old' value of the trait, and
#: the default trait value as the 'new' value. This allows other parts of the
#: traits package to recognize the assignment as the implicit default value
#: assignment, and treat it specially.
Uninitialized = _Uninitialized()

#-------------------------------------------------------------------------------
#  Singleton 'Undefined' object (used as undefined trait name and/or value):
#-------------------------------------------------------------------------------

Undefined = None

class _Undefined(object):
    """ Singleton 'Undefined' object (used as undefined trait name and/or value)
    """
    def __new__(cls):
        if Undefined is not None:
            return Undefined
        else:
            self = object.__new__(cls)
            return self

    def __repr__(self):
        return '<undefined>'

    def __reduce_ex__(self, protocol):
        return (_Undefined, ())

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return hash(type(self))

    def __ne__(self, other):
        return type(self) is not type(other)

#: Singleton object that indicates that a trait attribute has not yet had a
#: value set (i.e., its value is undefined). This object is used instead of
#: None, because None often has other meanings, such as that a value is not
#: used. When a trait attribute is first assigned a value, and its associated
#: trait notification handlers are called, Undefined is passed as the *old*
#: parameter, to indicate that the attribute previously had no value.
Undefined = _Undefined()

# Tell the C-base code about singleton 'Undefined' and 'Uninitialized' objects:
from . import ctraits
ctraits._undefined( Undefined, Uninitialized )

#-------------------------------------------------------------------------------
#  Singleton 'Missing' object (used as missing method argument marker):
#-------------------------------------------------------------------------------

class Missing ( object ):
    """ Singleton 'Missing' object (used as missing method argument marker).
    """
    def __repr__ ( self ):
        return '<missing>'

#: Singleton object that indicates that a method argument is missing from a
#: type-checked method signature.
Missing = Missing()

#-------------------------------------------------------------------------------
#  Singleton 'Self' object (used as object reference to current 'object'):
#-------------------------------------------------------------------------------

class Self ( object ):
    """ Singleton 'Self' object (used as object reference to current 'object').
    """
    def __repr__ ( self ):
        return '<self>'

#: Singleton object that references the current 'object'.
Self = Self()

#-------------------------------------------------------------------------------
#  Define a special 'string' coercion function:
#-------------------------------------------------------------------------------

def strx ( arg ):
    """ Wraps the built-in str() function to raise a TypeError if the
    argument is not of a type in StringTypes.
    """
    if isinstance( arg, StringTypes ):
       return str( arg )
    raise TypeError

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

StringTypes = ( str, unicode, int, long, float, complex )

#-------------------------------------------------------------------------------
#  Define a mapping of coercable types:
#-------------------------------------------------------------------------------

# Mapping of coercable types.
CoercableTypes = {
    long:    ( 11, long, int ),
    float:   ( 11, float, int ),
    complex: ( 11, complex, float, int ),
    unicode: ( 11, unicode, str )
}

#-------------------------------------------------------------------------------
#  Return a string containing the class name of an object with the correct
#  article (a or an) preceding it (e.g. 'an Image', 'a PlotValue'):
#-------------------------------------------------------------------------------

def class_of ( object ):
    """ Returns a string containing the class name of an object with the
    correct indefinite article ('a' or 'an') preceding it (e.g., 'an Image',
    'a PlotValue').
    """
    if isinstance( object, basestring ):
        return add_article( object )

    return add_article( object.__class__.__name__ )

#-------------------------------------------------------------------------------
#  Return a string containing the right article (i.e. 'a' or 'an') prefixed to
#  a specified string:
#-------------------------------------------------------------------------------

def add_article ( name ):
    """ Returns a string containing the correct indefinite article ('a' or 'an')
    prefixed to the specified string.
    """
    if name[:1].lower() in 'aeiou':
       return 'an ' + name

    return 'a ' + name

#----------------------------------------------------------------------------
#  Return a 'user-friendly' name for a specified trait:
#----------------------------------------------------------------------------

def user_name_for ( name ):
    """ Returns a "user-friendly" version of a string, with the first letter
    capitalized and with underscore characters replaced by spaces. For example,
    ``user_name_for('user_name_for')`` returns ``'User name for'``.
    """
    name       = name.replace( '_', ' ' )
    result     = ''
    last_lower = False

    for c in name:
        if c.isupper() and last_lower:
           result += ' '
        last_lower = c.islower()
        result    += c

    return result.capitalize()

#-------------------------------------------------------------------------------
#  Gets the path to the traits home directory:
#-------------------------------------------------------------------------------

_traits_home = None

def traits_home ( ):
    """ Gets the path to the Traits home directory.
    """
    global _traits_home

    if _traits_home is None:
        _traits_home = verify_path( join( ETSConfig.application_data,
                                          'traits' ) )

    return _traits_home

#-------------------------------------------------------------------------------
#  Verify that a specified path exists, and try to create it if it doesn't:
#-------------------------------------------------------------------------------

def verify_path ( path ):
    """ Verify that a specified path exists, and try to create it if it
        does not exist.
    """
    if not exists( path ):
        try:
            os.mkdir( path )
        except:
            pass

    return path

#-------------------------------------------------------------------------------
#  Returns the name of the module the caller's caller is located in:
#-------------------------------------------------------------------------------

def get_module_name ( level = 2 ):
    """ Returns the name of the module that the caller's caller is located in.
    """
    return sys._getframe( level ).f_globals.get( '__name__', '__main__' )

#-------------------------------------------------------------------------------
#  Returns a resource path calculated from the caller's stack:
#-------------------------------------------------------------------------------

def get_resource_path ( level = 2 ):
    """Returns a resource path calculated from the caller's stack.
    """
    module = sys._getframe( level ).f_globals.get( '__name__', '__main__' )

    path = None
    if module != '__main__':
        # Return the path to the module:
        try:
            path = dirname( getattr( sys.modules.get( module ), '__file__' ) )
        except:
            # Apparently 'module' is not a registered module...treat it like
            # '__main__':
            pass

    if path is None:
        # '__main__' is not a real module, so we need a work around:
        for path in [ dirname( sys.argv[0] ), getcwd() ]:
            if exists( path ):
                break

    # Handle application bundlers.  Since the python source files may be placed
    # in a zip file and therefore won't be directly accessable using standard
    # open/read commands, the app bundlers will look for resources (i.e.  data
    # files, images, etc.) in specific locations.  For py2app, this is in the
    # [myapp].app/Contents/Resources directory.  For py2exe, this is the same
    # directory as the [myapp].exe executable file generated by py2exe.
    frozen = getattr(sys, 'frozen', False)
    if frozen:
        if hasattr(sys, "_MEIPASS"):
            root = sys._MEIPASS
        elif frozen in ('macosx_app'):
            root = os.environ['RESOURCEPATH']
        elif frozen in ('dll', 'windows_exe', 'console_exe'):
            root = os.path.dirname(sys.executable)
        else:
            # Unknown app bundler, but try anyway
            root = os.path.dirname(sys.executable)
        if ".zip/" in path:
            zippath, image_path = path.split(".zip/")
            path = os.path.join(root, image_path)

    return path

#-------------------------------------------------------------------------------
#  Returns the value of an extended object attribute name of the form:
#  name[.name2[.name3...]]:
#-------------------------------------------------------------------------------

def xgetattr( object, xname, default = Undefined ):
    """ Returns the value of an extended object attribute name of the form:
        name[.name2[.name3...]].
    """
    names = xname.split( '.' )
    for name in names[:-1]:
        if default is Undefined:
            object = getattr( object, name )
        else:
            object = getattr( object, name, None )
            if object is None:
                return default

    if default is Undefined:
        return getattr( object, names[-1] )

    return getattr( object, names[-1], default )

#-------------------------------------------------------------------------------
#  Sets the value of an extended object attribute name of the form:
#  name[.name2[.name3...]]:
#-------------------------------------------------------------------------------

def xsetattr( object, xname, value ):
    """ Sets the value of an extended object attribute name of the form:
        name[.name2[.name3...]].
    """
    names = xname.split( '.' )
    for name in names[:-1]:
        object = getattr( object, name )

    setattr( object, names[-1], value )

#-------------------------------------------------------------------------------
#  Traits metadata selection functions:
#-------------------------------------------------------------------------------

def is_none ( value ):
    return (value is None)

def not_none ( value ):
    return (value is not None)

def not_false ( value ):
    return (value is not False)

def not_event ( value ):
    return (value != 'event')

def is_str ( value ):
    return isinstance( value, basestring )



vertex = """
varying vec4 worldCoords;
varying vec4 localCoords;
uniform mat3 transform;

void main()
{
    worldCoords = gl_ModelViewMatrix * gl_Vertex; 
    localCoords = gl_Vertex;
    gl_Position = ftransform();
}"""


radial = """
#line 15

uniform vec2 center;
uniform vec2 focalPoint;
uniform float radius;

uniform vec4 stops;

uniform vec4 stop0;
uniform vec4 stop1;
uniform vec4 stop2;
uniform vec4 stop3;
uniform vec4 stop4;

uniform float opacity;

uniform mat3 worldTransform;
uniform mat3 gradientTransform;
uniform mat3 invGradientTransform;

varying vec4 worldCoords;
varying vec4 localCoords; 

void main()
{
    vec4 result;
    
    vec3 tFocalPoint = invGradientTransform*vec3(focalPoint.x, focalPoint.y, 1.0);
    vec3 transformed = invGradientTransform*vec3(localCoords.x, localCoords.y, 1.0);

    float c = length(center - focalPoint);
    float alpha = atan((center.x-focalPoint.x),(center.y-focalPoint.y));
        
    float a = radius;
    float r = length(transformed.xy - focalPoint.xy);
    float theta = atan((transformed.x-focalPoint.x),(transformed.y-focalPoint.y));
    float cosAlphaTheta = cos(alpha-theta);
    float circR2 = sqrt(a*a + c*c*cosAlphaTheta*cosAlphaTheta - c*c) + c*cosAlphaTheta;

    float ratio = clamp(r/circR2, 0.0, 1.0);
            
    //calculate the intensity
    float intensity = ratio;

    if(intensity <= stops.x)
    {
        result.rgba = mix(stop0, stop1, (intensity / stops.x));
    }
    else if(intensity <= stops.y)
    {
        result.rgba = mix(stop1, stop2, (intensity-stops.x) / (stops.y-stops.x));
    }
    else if(intensity <= stops.z)
    {
        result.rgba = mix(stop2, stop3, (intensity-stops.y) / (stops.z-stops.y));
    }
    else if(intensity <= stops.w)
    {
        result.rgba = mix(stop3, stop4, (intensity-stops.z) / (stops.w-stops.z));
    }
    else
    {
        result.rgba = stop4;
    }

    result.a = result.a * opacity;

    gl_FragColor = result;
}"""

linear = """
#line 86

uniform vec2 start;
uniform vec2 end;

uniform float opacity;

uniform float canvasHeight;
uniform vec4 stops;

uniform vec4 stop0;
uniform vec4 stop1;
uniform vec4 stop2;
uniform vec4 stop3;
uniform vec4 stop4;

uniform mat3 worldTransform;
uniform mat3 gradientTransform;
uniform mat3 invGradientTransform;

varying vec4 worldCoords;
varying vec4 localCoords; 

void main()
{
    vec4 result;
            
    vec3 s = gradientTransform*vec3(start.x, start.y, 1);
    vec3 d = gradientTransform*vec3(end.x, end.y, 1); 
    vec3 l = localCoords.xyz;
        
    float num = (l.x - s.x)*(d.x - s.x) + (l.y - s.y)*(d.y - s.y);
    float denom = pow(abs(s.x - d.x), 2.0) + pow(abs(s.y - d.y), 2.0);
    float intensity =  clamp(num / denom, 0.0, 1.0);

    //calculate the intensity
    if(intensity <= stops.x)
    {
        result.rgba = mix(stop0, stop1, (intensity / stops.x));
    }
    else if(intensity <= stops.y)
    {
        result.rgba = mix(stop1, stop2, (intensity-stops.x) / (stops.y-stops.x));
    }
    else if(intensity <= stops.z)
    {
        result.rgba = mix(stop2, stop3, (intensity-stops.y) / (stops.z-stops.y));
    }
    else if(intensity <= stops.w)
    {
        result.rgba = mix(stop3, stop4, (intensity-stops.z) / (stops.w-stops.z));
    }
    else
    {
        result.rgba = stop4;
    }

    result.a = result.a * opacity;
    gl_FragColor = result;
}
"""


import wx
import wx.lib.buttons as buttons

from pyface.api import ImageResource


class GenToggleButtonEvent(wx.PyCommandEvent):
    """Event sent from the generic buttons when the button is activated. """
    def __init__(self, eventType, id, pressed, evt_obj):
        wx.PyCommandEvent.__init__(self, eventType, id)
        self.pressed = pressed
        self.evt_obj = evt_obj

    def Checked(self):
        return self.pressed

    IsChecked = Checked

    def GetEventObject(self):
        return self.evt_obj

        btn.SetToolTipString("Case sensitive match")


class FlatBitmapToggleButton(buttons.GenBitmapToggleButton):
    def __init__(self, parent, id, icon_name, pressed=False, tooltip_prefix="Toggle", *args, **kwargs):
        buttons.GenBitmapToggleButton.__init__(self, parent, -1, None, style=wx.BORDER_NONE)
        self.SetValue(pressed)
        self.tooltip_prefix = tooltip_prefix
        img = ImageResource(icon_name)
        bmp = img.create_bitmap()
        self.SetBitmapLabel(bmp)
        self.SetInitialSize()
        self.set_tooltip()

    def InitColours(self):
        buttons.GenBitmapToggleButton.InitColours(self)
        faceClr = self.GetBackgroundColour()
        r, g, b = faceClr.Get()
        fr, fg, fb = max(0,r-32), max(0,g-32), max(0,b-32)
        self.faceDnClr = wx.Colour(fr, fg, fb)

    def Notify(self):
        evt = GenToggleButtonEvent(wx.wxEVT_COMMAND_BUTTON_CLICKED, self.GetId(), not self.up, self)
        self.GetEventHandler().ProcessEvent(evt)
        self.set_tooltip()

    def set_tooltip(self):
        state = "off" if self.up else "on"
        if self.tooltip_prefix:
            state = "%s: %s" % (self.tooltip_prefix, state)
        self.SetToolTipString(state)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------

""" Table column object for Color traits.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from traitsui.table_column \
    import ObjectColumn

#-------------------------------------------------------------------------
#  'ColorColumn' class:
#-------------------------------------------------------------------------


class ColorColumn(ObjectColumn):
    """ Table column object for Color traits. """

#-- ObjectColumn Overrides -----------------------------------------------

    def get_cell_color(self, object):
        """ Returns the cell background color for the column for a specified
            object.
        """
        color_values = getattr(object, self.name + '_')
        if isinstance(color_values, tuple):
            tk_color = self._as_int_rgb_tuple(color_values)
        else:
            tk_color = super(ColorColumn, self).get_cell_color(object)
        return tk_color

    def get_value(self, object):
        """ Gets the value of the column for a specified object.
        """
        value = getattr(self.get_object(object), self.name)
        if isinstance(value, tuple):
            value = "(%3d, %3d, %3d)" % self._as_int_rgb_tuple(value[:-1])
        elif not isinstance(value, str):
            value = str(value)

        return value

#-- Private Methods ------------------------------------------------------

    def _as_int_rgb_tuple(self, color_values):
        """ Returns object color as RGB integers. """
        return (int(255 * color_values[0]),
                int(255 * color_values[1]),
                int(255 * color_values[2]))



#!/usr/bin/env python

import sys

from fs.opener import opener
from fs.commands.runner import Command
from fs.utils import print_fs
import errno


class FSServe(Command):

    usage = """fsserve [OPTION]... [PATH]
Serves the contents of PATH with one of a number of methods"""

    def get_optparse(self):
        optparse = super(FSServe, self).get_optparse()
        optparse.add_option('-t', '--type', dest='type', type="string", default="http",
                            help="Server type to create (http, rpc, sftp)", metavar="TYPE")
        optparse.add_option('-a', '--addr', dest='addr', type="string", default="127.0.0.1",
                            help="Server address", metavar="ADDR")
        optparse.add_option('-p', '--port', dest='port', type="int",
                            help="Port number", metavar="")
        return optparse

    def do_run(self, options, args):

        try:
            fs_url = args[0]
        except IndexError:
            fs_url = './'

        fs, path = self.open_fs(fs_url)

        if fs.isdir(path):
            fs = fs.opendir(path)
            path = '/'

        self.output("Opened %s\n" % fs, verbose=True)

        port = options.port

        try:

            if options.type == 'http':
                from fs.expose.http import serve_fs
                if port is None:
                    port = 80
                self.output("Starting http server on %s:%i\n" % (options.addr, port), verbose=True)
                serve_fs(fs, options.addr, port)

            elif options.type == 'rpc':
                from fs.expose.xmlrpc import RPCFSServer
                if port is None:
                    port = 80
                s = RPCFSServer(fs, (options.addr, port))
                self.output("Starting rpc server on %s:%i\n" % (options.addr, port), verbose=True)
                s.serve_forever()

            elif options.type == 'ftp':
                from fs.expose.ftp import serve_fs
                if port is None:
                    port = 21
                self.output("Starting ftp server on %s:%i\n" % (options.addr, port), verbose=True)
                serve_fs(fs, options.addr, port)

            elif options.type == 'sftp':
                from fs.expose.sftp import BaseSFTPServer
                import logging
                log = logging.getLogger('paramiko')
                if options.debug:
                    log.setLevel(logging.DEBUG)
                elif options.verbose:
                    log.setLevel(logging.INFO)
                ch = logging.StreamHandler()
                ch.setLevel(logging.DEBUG)
                log.addHandler(ch)

                if port is None:
                    port = 22
                server = BaseSFTPServer((options.addr, port), fs)
                try:
                    self.output("Starting sftp server on %s:%i\n" % (options.addr, port), verbose=True)
                    server.serve_forever()
                except Exception, e:
                    pass
                finally:
                    server.server_close()

            else:
                self.error("Server type '%s' not recognised\n" % options.type)

        except IOError, e:
            if e.errno == errno.EACCES:
                self.error('Permission denied\n')
                return 1
            else:
                self.error(str(e) + '\n')
                return 1

def run():
    return FSServe().run()

if __name__ == "__main__":
    sys.exit(run())



"""

  A version of functools.wraps for Python versions that don't support it.

Note that this module can't be named "functools" because it would shadow the
stdlib module that it tries to emulate.  Absolute imports would fix this
problem but are only availabe from Python 2.5.

"""

try:
    from functools import wraps as wraps
except ImportError:
    wraps = lambda f: lambda f: f



try:
    from fs.contrib.sqlitefs import SqliteFS
except ImportError:
    SqliteFS = None
from fs.tests import FSTestCases
import unittest

import os

if SqliteFS:
    class TestSqliteFS(unittest.TestCase, FSTestCases):
        def setUp(self):
            self.fs = SqliteFS("sqlitefs.db")            
        def tearDown(self):
            os.remove('sqlitefs.db')
        
        


#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The naming manager. """


# Enthought library imports.
from traits.api import HasTraits


class NamingManager(HasTraits):
    """ The naming manager. """

    ###########################################################################
    # 'NamingManager' interface.
    ###########################################################################

    def get_state_to_bind(self, obj, name, context):
        """ Returns the state of an object for binding.

        The naming manager asks the context for its list of STATE factories
        and then calls them one by one until it gets a non-None result
        indicating that the factory recognised the object and created state
        information for it.

        If none of the factories recognize the object (or if the context
        has no factories) then the object itself is returned.

        """

        # Local imports.
        from context import Context

        # We get the state factories from the context's environment.
        state_factories = context.environment[Context.STATE_FACTORIES]

        for state_factory in state_factories:
            state = state_factory.get_state_to_bind(obj, name, context)
            if state is not None:
                break

        else:
            state = obj

        return state

    def get_object_instance(self, info, name, context):
        """ Creates an object using the specified state information.

        The naming manager asks the context for its list of OBJECT factories
        and calls them one by one until it gets a non-None result, indicating
        that the factory recognised the information and created an object.

        If none of the factories recognize the state information (or if the
        context has no factories) then the state information itself is
        returned.

        """

        # Local imports.
        from context import Context

        # We get the object factories from the context's environment.
        object_factories = context.environment[Context.OBJECT_FACTORIES]

        for object_factory in object_factories:
            obj = object_factory.get_object_instance(info, name, context)
            if obj is not None:
                break

        else:
            obj = info

        return obj


# Singleton instance.
naming_manager = NamingManager()

### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The node type for NON-contexts in a naming system. """


# Enthought library imports.
from apptools.naming.api import Context
from pyface.tree.api import NodeType


class ObjectNodeType(NodeType):
    """ The node type for NON-contexts in a naming system. """

    ###########################################################################
    # 'NodeType' interface.
    ###########################################################################

    # 'node' in this case is a 'Binding' instance.
    def is_type_for(self, node):
        """ Returns True if this node type recognized a node. """

        return True

    def allows_children(self, node):
        """ Does the node allow children (ie. a folder vs a file). """

        return False

    def get_drag_value(self, node):
        """ Get the value that is dragged for a node. """

        return node.obj

    def is_editable(self, node):
        """ Returns True if the node is editable, otherwise False.

        If the node is editable, its text can be set via the UI.

        """

        return True

    def get_text(self, node):
        """ Returns the label text for a node. """

        return node.name

    def can_set_text(self, node, text):
        """ Returns True if the node's label can be set. """

        # The parent context will NOT be None here (an object is ALWAYS
        # contained in a context).
        parent = node.context

        return len(text.strip()) > 0 and text not in parent.list_names('')

    def set_text(self, node, text):
        """ Sets the label text for a node. """

        node.context.rename(node.name, text)
        node.name = text

        return

##### EOF #####################################################################



""" The interface for perspective items. """


# Enthought library imports.
from traits.api import Enum, Float, Interface, Str


class IPerspectiveItem(Interface):
    """ The interface for perspective items. """

    # The Id of the view to display in the perspective.
    id = Str

    # The position of the view relative to the item specified in the
    # 'relative_to' trait.
    #
    # 'top'    puts the view above the 'relative_to' item.
    # 'bottom' puts the view below the 'relative_to' item.
    # 'left'   puts the view to the left of  the 'relative_to' item.
    # 'right'  puts the view to the right of the 'relative_to' item.
    # 'with'   puts the view in the same region as the 'relative_to' item.
    #
    # If the position is specified as 'with' you must specify a 'relative_to'
    # item other than the editor area (i.e., you cannot position a view 'with'
    # the editor area).
    position = Enum('left', 'top', 'bottom', 'right', 'with')

    # The Id of the view to position relative to. If this is not specified
    # (or if no view exists with this Id) then the view will be placed relative
    # to the editor area.
    relative_to = Str

    # The width of the item (as a fraction of the window width).
    #
    # e.g. 0.5 == half the window width.
    #
    # Note that this is treated as a suggestion, and it may not be possible
    # for the workbench to allocate the space requested.
    width = Float(-1)

    # The height of the item (as a fraction of the window height).
    #
    # e.g. 0.5 == half the window height.
    #
    # Note that this is treated as a suggestion, and it may not be possible
    # for the workbench to allocate the space requested.
    height = Float(-1)

    # The style of the dock window.
    style_hint = Enum('tab', 'horizontal', 'vertical', 'fixed')

#### EOF ######################################################################



import wx
from wx.lib.agw.rulerctrl import RulerCtrl
import wx.lib.scrolledpanel as scrolled


class ZoomRuler(scrolled.ScrolledPanel):
    """Zoomable ruler that uses a scrollbar and resize to implement the zoom.
    """
    def __init__(self, parent, **kwargs):
        scrolled.ScrolledPanel.__init__(self, parent, -1, size=(-1, 50), style=wx.HSCROLL)
        self.ruler = RulerCtrl(self, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.ruler, 1, wx.EXPAND, 0)
        self.SetSizer(sizer)
        self.ShowScrollbars(wx.SHOW_SB_ALWAYS, wx.SHOW_SB_NEVER)
        self.SetupScrolling(scroll_y=False)
        sizer.Layout()
        self.Fit()

        size = (1000,40)
        self.zoom_parent(size)

        #self.panel.Bind(wx.EVT_SIZE, self.panel_size)
        self.ruler.Bind(wx.EVT_MOUSE_EVENTS, self.on_mouse_events)

    def on_mouse_events(self, event):
        """Overriding the ruler to capture wheel events
        """
        wheel_dir = event.GetWheelRotation()
        if wheel_dir:
            if event.ControlDown():
                if wheel_dir < 0:
                    self.zoom_out()
                elif wheel_dir > 0:
                    self.zoom_in()
            event.Skip()
        else:
            self.ruler.OnMouseEvents(event)

    def zoom_out(self):
        size = self.ruler.GetSize()
        newsize = size - (50, 0)
        self.zoom_parent(newsize)

    def zoom_in(self):
        size = self.ruler.GetSize()
        newsize = size + (50, 0)
        self.zoom_parent(newsize)

    def zoom_parent(self, size):
        self.ruler.SetSize(size)
        self.SetVirtualSize(size)


if __name__ == "__main__":
    import wx.lib.scrolledpanel as scrolled

    app = wx.PySimpleApp()
    frm = wx.Frame(None,-1,"Test",style=wx.TAB_TRAVERSAL|wx.DEFAULT_FRAME_STYLE,
                   size=(800,400))
    panel = wx.Panel(frm)
    sizer = wx.BoxSizer(wx.VERTICAL)

    text = wx.StaticText(panel, -1, "Just a placeholder here.")
    sizer.Add(text, 1, wx.EXPAND)

    scroll = ZoomRuler(panel)
    sizer.Add(scroll, 0, wx.EXPAND)

    panel.SetAutoLayout(True)
    panel.SetSizer(sizer)
    #sizer.Fit(panel)
    #sizer.SetSizeHints(panel)
    panel.Layout()
    app.SetTopWindow(frm)
    frm.Show()
    app.MainLoop()



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

""" Tests the state function functionality of the apptools.sweet_pickle
    framework.
"""

# Standard library imports.
import unittest
import logging

# Enthought library imports
import apptools.sweet_pickle as sweet_pickle
from apptools.sweet_pickle.global_registry import _clear_global_registry
from traits.api import Bool, Float, HasTraits, Int, Str


logger = logging.getLogger(__name__)


##############################################################################
# Classes to use within the tests
##############################################################################

# Need complete package name so that mapping matches correctly.
# The problem here is the Python loader that will load the same module with
# multiple names in sys.modules due to relative naming. Nice.
from apptools.sweet_pickle.tests.state_function_classes import Foo, Bar, Baz

##############################################################################
# State functions to use within the tests
##############################################################################

def bar_state_function(state):
    for old, new in [('b1', 'b2'), ('f1', 'f2'), ('i1', 'i2'), ('s1', 's2')]:
        state[new] = state[old]
        del state[old]
    state['_enthought_pickle_version'] = 2
    return state


##############################################################################
# class 'StateFunctionTestCase'
##############################################################################

class StateFunctionTestCase(unittest.TestCase):
    """ Tests the state function functionality of the apptools.sweet_pickle
        framework.
    """

    ##########################################################################
    # 'TestCase' interface
    ##########################################################################

    ### public interface #####################################################

    def setUp(self):
        """ Creates the test fixture.

            Overridden here to ensure each test starts with an empty global
            registry.
        """
        # Clear the global registry
        _clear_global_registry()

        # Cache a reference to the new global registry
        self.registry = sweet_pickle.get_global_registry()

        # Add the class mappings to the registry
        self.registry.add_mapping_to_class(Foo.__module__, Foo.__name__,
            Bar)
        self.registry.add_mapping_to_class(Bar.__module__, Bar.__name__,
            Baz)


    ##########################################################################
    # 'StateFunctionTestCase' interface
    ##########################################################################

    ### public interface #####################################################

    def test_normal_setstate(self):
        """ Validates that only existing setstate methods are called when
            there are no registered state functions in the class chain.
        """
        # Validate that unpickling the first class gives us an instance of
        # the third class with the appropriate attribute values.  It will have
        # the default Foo values (because there is no state function to move
        # them) and also the default Baz values (since they inherit the
        # trait defaults because nothing overwrote the values.)
        start = Foo()
        end = sweet_pickle.loads(sweet_pickle.dumps(start))
        self.assertEqual(True, isinstance(end, Baz))
        self._assertAttributes(end, 1, (False, 1, 1, 'foo'))
        self._assertAttributes(end, 2, None)
        self._assertAttributes(end, 3, (False, 3, 3, 'baz'))

        # Validate that unpickling the second class gives us an instance of
        # the third class with the appropriate attribute values.  It will have
        # only the Baz attributes with the Bar values (since the __setstate__
        # on Baz converted the Bar attributes to Baz attributes.)
        start = Bar()
        end = sweet_pickle.loads(sweet_pickle.dumps(start))
        self.assertEqual(True, isinstance(end, Baz))
        self._assertAttributes(end, 2, None)
        self._assertAttributes(end, 3, (True, 2, 2, 'bar'))


    def test_unpickled_chain_functionality(self):
        """ Validates that the registered state functions are used when
            unpickling.
        """
        # Add the state function to the registry
        self.registry.add_state_function_for_class(Bar, 2,
            bar_state_function)

        # Validate that unpickling the first class gives us an instance of
        # the third class with the appropriate attribute values.
        start = Foo()
        end = sweet_pickle.loads(sweet_pickle.dumps(start))
        self.assertEqual(True, isinstance(end, Baz))
        self._assertAttributes(end, 1, None)
        self._assertAttributes(end, 2, None)
        self._assertAttributes(end, 3, (False, 1, 1, 'foo'))

        # Validate that unpickling the second class gives us an instance of
        # the third class.
        start = Bar()
        end = sweet_pickle.loads(sweet_pickle.dumps(start))
        self.assertEqual(True, isinstance(end, Baz))
        self._assertAttributes(end, 2, None)
        self._assertAttributes(end, 3, (True, 2, 2, 'bar'))


    ### protected interface ##################################################

    def _assertAttributes(self, obj, suffix, values):
        """ Ensures that the specified object's attributes with the specified
            suffix have the expected values.  If values is None, then the
            attributes shouldn't exist.
        """
        attributeNames = ['b', 'f', 'i', 's']
        for i in range(len(attributeNames)):
            name = attributeNames[i] + str(suffix)
            if values is None:
                self.assertEqual(False, hasattr(obj, name),
                    'Obj [%s] has attribute [%s]' % (obj, name))
            else:
                self.assertEqual(values[i], getattr(obj, name),
                    'Obj [%s] attribute [%s] has [%s] instead of [%s]' % \
                    (obj, name, values[i], getattr(obj, name)))


if __name__ == "__main__":
    unittest.main()


### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" Enthought pyface package component
"""

# Major package imports.
import wx.stc

# Enthought library imports.
from traits.api import Bool, Event, provides, Unicode

# Local imports.
from pyface.i_python_editor import IPythonEditor, MPythonEditor
from pyface.key_pressed_event import KeyPressedEvent
from pyface.wx.python_stc import PythonSTC, faces
from .widget import Widget


@provides(IPythonEditor)
class PythonEditor(MPythonEditor, Widget):
    """ The toolkit specific implementation of a PythonEditor.  See the
    IPythonEditor interface for the API documentation.
    """


    #### 'IPythonEditor' interface ############################################

    dirty = Bool(False)

    path = Unicode

    show_line_numbers = Bool(True)

    #### Events ####

    changed = Event

    key_pressed = Event(KeyPressedEvent)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, parent, **traits):
        """ Creates a new pager. """

        # Base class constructor.
        super(PythonEditor, self).__init__(**traits)

        # Create the toolkit-specific control that represents the widget.
        self.control = self._create_control(parent)

        return

    ###########################################################################
    # 'PythonEditor' interface.
    ###########################################################################

    def load(self, path=None):
        """ Loads the contents of the editor. """

        if path is None:
            path = self.path

        # We will have no path for a new script.
        if len(path) > 0:
            f = open(self.path, 'r')
            text = f.read()
            f.close()

        else:
            text = ''

        self.control.SetText(text)
        self.dirty = False

        return

    def save(self, path=None):
        """ Saves the contents of the editor. """

        if path is None:
            path = self.path

        f = open(path, 'w')
        f.write(self.control.GetText())
        f.close()

        self.dirty = False

        return

    def set_style(self, n, fore, back):

        self.control.StyleSetForeground(n, fore)
        #self.StyleSetBackground(n, '#c0c0c0')
        #self.StyleSetBackground(n, '#ffffff')
        self.control.StyleSetBackground(n, back)
        self.control.StyleSetFaceName(n, "courier new")
        self.control.StyleSetSize(n, faces['size'])

        #self.StyleSetForeground(n, "#f0f0f0")
        ##self.StyleSetBackground(n, "#000000")
        #self.StyleSetFaceName(n, "courier new")
        #self.StyleSetSize(n, 20)
        #self.StyleSetUnderline(n, 1)
        #self.StyleSetItalic(n, 1)
        #self.StyleSetBold(n, 1)
        #StyleClearAll
        #StyleResetDefault
        #StyleSetCase
        #StyleSetChangeable
        #StyleSetCharacterSet
        #StyleSetEOLFilled
        #StyleSetFont
        #StyleSetFontAttr
        #StyleSetHotSpot
        #StyleSetSpec --- batch
        #StyleSetVisible

        return

    def select_line(self, lineno):
        """ Selects the specified line. """

        start = self.control.PositionFromLine(lineno)
        end   = self.control.GetLineEndPosition(lineno)

        self.control.SetSelection(start, end)

        return

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _path_changed(self):
        """ Handle a change to path. """

        self._changed_path()

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_control(self, parent):
        """ Creates the toolkit-specific control for the widget. """

        # Base-class constructor.
        self.control = stc = PythonSTC(parent, -1)

        # No folding!
        stc.SetProperty("fold", "0")

        # Mark the maximum line size.
        stc.SetEdgeMode(wx.stc.STC_EDGE_LINE)
        stc.SetEdgeColumn(79)

        # Display line numbers in the margin.
        if self.show_line_numbers:
            stc.SetMarginType(1, wx.stc.STC_MARGIN_NUMBER)
            stc.SetMarginWidth(1, 45)
            self.set_style(wx.stc.STC_STYLE_LINENUMBER, "#000000", "#c0c0c0")
        else:
            stc.SetMarginWidth(1, 4)
            self.set_style(wx.stc.STC_STYLE_LINENUMBER, "#ffffff", "#ffffff")

        # Create 'tabs' out of spaces!
        stc.SetUseTabs(False)

        # One 'tab' is 4 spaces.
        stc.SetIndent(4)

        # Line ending mode.
        stc.SetEOLMode(wx.stc.STC_EOL_LF) # Unix
        #self.SetEOLMode(wx.stc.STC_EOL_CR) # Apple Mac
        #self.SetEOLMode(wx.stc.STC_EOL_CRLF) # Windows

        ##########################################
        # Global styles for all languages.
        ##########################################

        self.set_style(wx.stc.STC_STYLE_DEFAULT, "#000000", "#ffffff")
        self.set_style(wx.stc.STC_STYLE_CONTROLCHAR, "#000000", "#ffffff")
        self.set_style(wx.stc.STC_STYLE_BRACELIGHT, "#000000", "#ffffff")
        self.set_style(wx.stc.STC_STYLE_BRACEBAD, "#000000", "#ffffff")

        ##########################################
        # Python styles.
        ##########################################

        # White space
        self.set_style(wx.stc.STC_P_DEFAULT, "#000000", "#ffffff")

        # Comment
        self.set_style(wx.stc.STC_P_COMMENTLINE, "#007f00", "#ffffff")

        # Number
        self.set_style(wx.stc.STC_P_NUMBER, "#007f7f", "#ffffff")

        # String
        self.set_style(wx.stc.STC_P_STRING, "#7f007f", "#ffffff")

        # Single quoted string
        self.set_style(wx.stc.STC_P_CHARACTER, "#7f007f", "#ffffff")

        # Keyword
        self.set_style(wx.stc.STC_P_WORD, "#00007f", "#ffffff")

        # Triple quotes
        self.set_style(wx.stc.STC_P_TRIPLE, "#7f0000", "#ffffff")

        # Triple double quotes
        self.set_style(wx.stc.STC_P_TRIPLEDOUBLE, "#ff0000", "#ffffff")

        # Class name definition
        self.set_style(wx.stc.STC_P_CLASSNAME, "#0000ff", "#ffffff")

        # Function or method name definition
        self.set_style(wx.stc.STC_P_DEFNAME, "#007f7f", "#ffffff")

        # Operators
        self.set_style(wx.stc.STC_P_OPERATOR, "#000000", "#ffffff")

        # Identifiers
        self.set_style(wx.stc.STC_P_IDENTIFIER, "#000000", "#ffffff")

        # Comment-blocks
        self.set_style(wx.stc.STC_P_COMMENTBLOCK, "#007f00", "#ffffff")

        # End of line where string is not closed
        self.set_style(wx.stc.STC_P_STRINGEOL, "#000000", "#ffffff")

        ##########################################
        # Events.
        ##########################################

        # By default, the will fire EVT_STC_CHANGE evented for all mask values
        # (STC_MODEVENTMASKALL). This generates too many events.
        stc.SetModEventMask(wx.stc.STC_MOD_INSERTTEXT |
                            wx.stc.STC_MOD_DELETETEXT |
                            wx.stc.STC_PERFORMED_UNDO |
                            wx.stc.STC_PERFORMED_REDO)

        # Listen for changes to the file.
        wx.stc.EVT_STC_CHANGE(stc, stc.GetId(), self._on_stc_changed)

        # Listen for key press events.
        wx.EVT_CHAR(stc, self._on_char)

        # Load the editor's contents.
        self.load()

        return stc

    #### wx event handlers ####################################################

    def _on_stc_changed(self, event):
        """ Called whenever a change is made to the text of the document. """

        self.dirty = True
        self.changed = True

        # Give other event handlers a chance.
        event.Skip()

        return

    def _on_char(self, event):
        """ Called whenever a change is made to the text of the document. """

        self.key_pressed = KeyPressedEvent(
            alt_down     = event.m_altDown == 1,
            control_down = event.m_controlDown == 1,
            shift_down   = event.m_shiftDown == 1,
            key_code     = event.m_keyCode,
            event        = event
        )

        # Give other event handlers a chance.
        event.Skip()

        return

#### EOF ######################################################################



#-------------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   02/05/2007
#
#-------------------------------------------------------------------------------

""" Pyface 'FeatureBar' support.

    Defines the 'FeatureBar' class which displays and allows the user to
    interact with a set of DockWindowFeatures for a specified DockControl.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from traits.api \
    import HasPrivateTraits, Instance, Bool, Event, Color

from pyface.wx.drag_and_drop \
    import PythonDropTarget, PythonDropSource

from dock_sizer \
    import DockControl, FEATURE_EXTERNAL_DRAG

from ifeature_tool \
    import IFeatureTool

#-------------------------------------------------------------------------------
#  'FeatureBar' class:
#-------------------------------------------------------------------------------

class FeatureBar ( HasPrivateTraits ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # The wx.Window which is the parent for the FeatureBar:
    parent = Instance( wx.Window )

    # The DockControl whose features are being displayed:
    dock_control = Instance( DockControl )

    # The wx.Window being used for the FeatureBar:
    control = Instance( wx.Window )

    # Event posted when the user has completed using the FeatureBar:
    completed = Event

    # The background color for the FeatureBar:
    bg_color = Color( 0xDBEEF7, allow_none = True )

    # The border color for the FeatureBar:
    border_color = Color( 0X2583AF, allow_none = True )

    # Should the feature bar display horizontally (or vertically)?
    horizontal = Bool( True )

    #---------------------------------------------------------------------------
    #  Hides the feature bar:
    #---------------------------------------------------------------------------

    def hide ( self ):
        """ Hides the feature bar.
        """
        if self.control is not None:
            self.control.Hide()

    #---------------------------------------------------------------------------
    #  Shows the feature bar:
    #---------------------------------------------------------------------------

    def show ( self ):
        """ Shows the feature bar.
        """
        # Make sure all prerequisites are met:
        dock_control, parent = self.dock_control, self.parent
        if (dock_control is None) or (parent is None):
            return

        # Create the actual control (if needed):
        control = self.control
        if control is None:
            self.control = control = wx.Frame( None, -1, '',
                                               style = wx.BORDER_NONE )

            # Set up the 'erase background' event handler:
            wx.EVT_ERASE_BACKGROUND( control, self._erase_background )

            # Set up the 'paint' event handler:
            wx.EVT_PAINT( control, self._paint )

            # Set up mouse event handlers:
            wx.EVT_LEFT_DOWN(    control, self._left_down )
            wx.EVT_LEFT_UP(      control, self._left_up )
            wx.EVT_RIGHT_DOWN(   control, self._right_down )
            wx.EVT_RIGHT_UP(     control, self._right_up )
            wx.EVT_MOTION(       control, self._mouse_move )
            wx.EVT_ENTER_WINDOW( control, self._mouse_enter )

            control.SetDropTarget( PythonDropTarget( self ) )

        # Calculate the best size and position for the feature bar:
        size       = wx.Size( 32, 32 )
        width      = height = 0
        horizontal = self.horizontal
        for feature in dock_control.active_features:
            bitmap = feature.bitmap
            if bitmap is not None:
                if horizontal:
                    width += (bitmap.GetWidth() + 3)
                    height = max( height, bitmap.GetHeight() )
                else:
                    width   = max( width, bitmap.GetWidth() )
                    height += (bitmap.GetHeight() + 3)

        if width > 0:
            if horizontal:
                size = wx.Size( width + 5, height + 8 )
            else:
                size = wx.Size( width + 8, height + 5 )

        control.SetSize( size )
        px, py = parent.GetScreenPosition()
        fx, fy = dock_control.feature_popup_position
        control.SetPosition( wx.Point( px + fx, py + fy ) )
        control.Show()

    #-- Window Event Handlers --------------------------------------------------

    #---------------------------------------------------------------------------
    #  Handles repainting the window:
    #---------------------------------------------------------------------------

    def _paint ( self, event ):
        """ Handles repainting the window.
        """
        window = self.control
        dx, dy = window.GetSizeTuple()
        dc     = wx.PaintDC( window )

        # Draw the feature container:
        bg_color     = self.bg_color
        border_color = self.border_color
        if (bg_color is not None) or (border_color is not None):
            if border_color is None:
                dc.SetPen( wx.TRANSPARENT_PEN )
            else:
                dc.SetPen( wx.Pen( border_color, 1, wx.SOLID ) )
            if bg_color is None:
                dc.SetBrush( wx.TRANSPARENT_PEN )
            else:
                dc.SetBrush( wx.Brush( bg_color, wx.SOLID ) )
            dc.DrawRectangle( 0, 0, dx, dy )

        # Draw the feature icons:
        if self.horizontal:
            x = 4
            for feature in self.dock_control.active_features:
                bitmap = feature.bitmap
                if bitmap is not None:
                    dc.DrawBitmap( bitmap, x, 4, True )
                    x += (bitmap.GetWidth() + 3)
        else:
            y = 4
            for feature in self.dock_control.active_features:
                bitmap = feature.bitmap
                if bitmap is not None:
                    dc.DrawBitmap( bitmap, 4, y, True )
                    y += (bitmap.GetHeight() + 3)

    #---------------------------------------------------------------------------
    #  Handles erasing the window background:
    #---------------------------------------------------------------------------

    def _erase_background ( self, event ):
        """ Handles erasing the window background.
        """
        pass

    #---------------------------------------------------------------------------
    #  Handles the left mouse button being pressed:
    #---------------------------------------------------------------------------

    def _left_down ( self, event ):
        """ Handles the left mouse button being pressed.
        """
        self._feature  = self._feature_at( event )
        self._dragging = False
        self._xy       = ( event.GetX(), event.GetY() )
        #self.control.CaptureMouse()

    #---------------------------------------------------------------------------
    #  Handles the left mouse button being released:
    #---------------------------------------------------------------------------

    def _left_up ( self, event ):
        """ Handles the left mouse button being released.
        """
        #self.control.ReleaseMouse()
        self._dragging         = None
        feature, self._feature = self._feature, None
        if feature is not None:
            if feature is self._feature_at( event ):
                self.control.ReleaseMouse()
                self.completed = True
                feature._set_event( event )
                feature.click()

    #---------------------------------------------------------------------------
    #  Handles the right mouse button being pressed:
    #---------------------------------------------------------------------------

    def _right_down ( self, event ):
        """ Handles the right mouse button being pressed.
        """
        self._feature  = self._feature_at( event )
        self._dragging = False
        self._xy       = ( event.GetX(), event.GetY() )
        #self.control.CaptureMouse()

    #---------------------------------------------------------------------------
    #  Handles the right mouse button being released:
    #---------------------------------------------------------------------------

    def _right_up ( self, event ):
        """ Handles the right mouse button being released.
        """
        #self.control.ReleaseMouse()
        self._dragging         = None
        feature, self._feature = self._feature, None
        if feature is not None:
            if feature is self._feature_at( event ):
                self.control.ReleaseMouse()
                self.completed = True
                feature._set_event( event )
                feature.right_click()

    #---------------------------------------------------------------------------
    #  Handles the mouse moving over the window:
    #---------------------------------------------------------------------------

    def _mouse_move ( self, event ):
        """ Handles the mouse moving over the window.
        """
        # Update tooltips if no mouse button is currently pressed:
        if self._dragging is None:
            feature = self._feature_at( event )
            if feature is not self._tooltip_feature:
                self._tooltip_feature = feature
                tooltip = ''
                if feature is not None:
                    tooltip = feature.tooltip
                wx.ToolTip.Enable( False )
                wx.ToolTip.Enable( True )
                self.control.SetToolTip( wx.ToolTip( tooltip ) )

            # Check to see if the mouse has left the window, and mark it
            # completed if it has:
            x, y   = event.GetX(), event.GetY()
            dx, dy = self.control.GetSizeTuple()
            if (x < 0) or (y < 0) or (x >= dx) or (y >= dy):
                self.control.ReleaseMouse()
                self._tooltip_feature = None
                self.completed        = True

            return

        # Check to see if we are in 'drag mode' yet:
        if not self._dragging:
            x, y = self._xy
            if (abs( x - event.GetX() ) + abs( y - event.GetY() )) < 3:
                return

            self._dragging = True

            # Check to see if user is trying to drag a 'feature':
            feature = self._feature
            if feature is not None:
                feature._set_event( event )

                prefix = button = ''
                if event.RightIsDown():
                    button = 'right_'
                if event.ControlDown():
                    prefix = 'control_'
                elif event.AltDown():
                    prefix = 'alt_'
                elif event.ShiftDown():
                    prefix = 'shift_'

                object = getattr( feature, '%s%sdrag' % ( prefix, button ) )()
                if object is not None:
                    self.control.ReleaseMouse()
                    self._feature  = None
                    self.completed = True
                    self.dock_control.pre_drag_all( object )
                    PythonDropSource( self.control, object )
                    self.dock_control.post_drag_all()
                    self._dragging = None

    #---------------------------------------------------------------------------
    #  Handles the mouse entering the window:
    #---------------------------------------------------------------------------

    def _mouse_enter ( self, event ):
        """ Handles the mouse entering the window.
        """
        self.control.CaptureMouse()

#-- Drag and drop event handlers: ----------------------------------------------

    #---------------------------------------------------------------------------
    #  Handles a Python object being dropped on the control:
    #---------------------------------------------------------------------------

    def wx_dropped_on ( self, x, y, data, drag_result ):
        """ Handles a Python object being dropped on the window.
        """
        # Determine what, if any, feature the object was dropped on:
        feature = self._can_drop_on_feature( x, y, data )

        # Indicate use of the feature bar is complete:
        self.completed = True

        # Reset any drag state information:
        self.dock_control.post_drag( FEATURE_EXTERNAL_DRAG )

        # Check to see if the data was dropped on a feature or not:
        if feature is not None:
            if isinstance( data, IFeatureTool ):
                # Handle an object implementing IFeatureTool being dropped:
                dock_control = feature.dock_control
                data.feature_dropped_on_dock_control( dock_control )
                data.feature_dropped_on( dock_control.object )
            else:
                # Handle a normal object being dropped:
                wx, wy = self.control.GetScreenPosition()
                feature.set( x = wx + x, y = wy + y )
                feature.drop( data )

            return drag_result

        return wx.DragNone

    #---------------------------------------------------------------------------
    #  Handles a Python object being dragged over the control:
    #---------------------------------------------------------------------------

    def wx_drag_over ( self, x, y, data, drag_result ):
        """ Handles a Python object being dragged over the control.
        """
        # Handle the case of dragging a normal object over a 'feature':
        if self._can_drop_on_feature( x, y, data ) is not None:
            return drag_result

        return wx.DragNone

    #---------------------------------------------------------------------------
    #  Handles a dragged Python object leaving the window:
    #---------------------------------------------------------------------------

    def wx_drag_leave ( self, data ):
        """ Handles a dragged Python object leaving the window.
        """
        # Indicate use of the feature bar is complete:
        self.completed = True

        # Reset any drag state information:
        self.dock_control.post_drag( FEATURE_EXTERNAL_DRAG )

    #-- Private Methods --------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Returns a feature that the pointer is over and which can accept the
    #  specified data:
    #---------------------------------------------------------------------------

    def _can_drop_on_feature ( self, x, y, data ):
        """ Returns a feature that the pointer is over and which can accept the
            specified data.
        """
        feature = self._feature_at( FakeEvent( x, y ) )
        if (feature is not None) and feature.can_drop( data ):
            return feature

        return None

    #---------------------------------------------------------------------------
    #  Returns the DockWindowFeature (if any) at a specified window position:
    #---------------------------------------------------------------------------

    def _feature_at ( self, event ):
        """ Returns the DockWindowFeature (if any) at a specified window
            position.
        """
        if self.horizontal:
            x = 4
            for feature in self.dock_control.active_features:
                bitmap = feature.bitmap
                if bitmap is not None:
                    bdx = bitmap.GetWidth()
                    if self._is_in( event, x, 4, bdx, bitmap.GetHeight() ):
                        return feature

                    x += (bdx + 3)
        else:
            y = 4
            for feature in self.dock_control.active_features:
                bitmap = feature.bitmap
                if bitmap is not None:
                    bdy = bitmap.GetHeight()
                    if self._is_in( event, 4, y, bitmap.GetWidth(), bdy ):
                        return feature

                    y += (bdy + 3)

        return None

    #---------------------------------------------------------------------------
    #  Returns whether or not an event is within a specified bounds:
    #---------------------------------------------------------------------------

    def _is_in ( self, event, x, y, dx, dy ):
        """ Returns whether or not an event is within a specified bounds.
        """
        return ((x <= event.GetX() < (x + dx)) and
                (y <= event.GetY() < (y + dy)))

#-------------------------------------------------------------------------------
#  'FakeEvent' class:
#-------------------------------------------------------------------------------

class FakeEvent ( object ):

    def __init__ ( self, x, y ):
        self.x, self.y = x, y

    def GetX ( self ): return self.x
    def GetY ( self ): return self.y




#!/usr/bin/env python

# for py2/3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals


"""
Module for manipulating netcdf (and other) particle files

This is useful for working with particle file output from GNOME,
and is a test case for working with what hopefully will be a
CF standard (or SOME standard..)

"""

from __future__ import division
# from datetime import datetime

import numpy as np

import netCDF4

# default attributes -- can be updated by user later.

file_attributes = {'conventions': "CF-1.6",
                   'title': "Sample data/file for particle trajectory format",
                   'institution': "NOAA Emergency Response Division",
                   'source': "example data from nc_particles",
                   'history': "Evolved with discussion on CF-metadata listserve",
                   'references': '',
                   'comment': 'Some simple test data',
                   'CF:featureType': "particle_trajectory" ,
                   }

# variable attributes for the standard variables
# these will be used when the data are created.
# you can use a different set if you want.

var_attributes = {'time': {'long_name': 'time since the beginning of the simulation',
                           'standard_name': 'time',
                           'calendar': 'gregorian',
                           'standard_name': 'time',
                           'comment': 'unspecified time zone',
                           # units will get set based on data
                           },
                  'particle_count': {'units': '1',
                                     'long_name': 'number of particles in a given timestep',
                                     'ragged_row_count': 'particle count at nth timestep',
                                     },
                  'longitude': {'long_name': 'longitude of the particle',
                                'standard_name': 'longitude',
                                'units': 'degrees_east',
                                },
                  'latitude': {'long_name': 'latitude of the particle',
                               'standard_name': 'latitude',
                               'units': 'degrees_north',
                               },
                  'depth': {'long_name': 'particle depth below sea surface',
                            'standard_name': 'depth',
                            'units': 'meters',
                            'axis': 'z positive down',
                            },
                  'mass': {'long_name': 'mass of particle',
                           'units': 'grams',
                           },
                  'age': {'long_name': 'age of particle from time of release',
                          'units': 'seconds',
                          },
                  'status_code': {'long_name': 'particle status code',
                                  'valid_range': (0, 5),
                                  'flag_values': "7 12 0 10 2 3",
                                  'flag_meanings': "0: not_released, 2: in_water, 3: on_land, 7: off_maps, 10: evaporated, 12: to_be_removed,"
                                  },
                  'id': {'long_name': 'particle ID',
                         },
                  }

# alternate names:
var_attributes['lon'] = var_attributes['longitude']
var_attributes['lat'] = var_attributes['latitude']

# variables used to support the stucture of the file, rather than data
# used to remove them from the list of available data variables
SPECIAL_VARIABLES = ['time', 'particle_count']


class Writer(object):
    def __init__(self,
                 filename,
                 num_timesteps=None,
                 ref_time=None,
                 file_attributes=file_attributes,
                 var_attributes=var_attributes,
                 nc_version=4,
                 ):

        """
        create a nc_particle file Writer

        Creates the netccdf file, opens it for writing,
        writes the global attributes, creates required variables etc.

        :param filename: name of netcdf file to open - if it exists,
                         it will be written over!

        :param num_timesteps=None: number of timesteps that will be output. Must be
                                   defined for netcdf3. Can be None for netcdf4
        :type num_timesteps: integer

        :param ref_time=None: reference time for time units (i.e. seconds since..).
                              If None, the first time used will be used.
        :type ref_time: datetime object

        :param file_attributes: keys and values for teh file-level attributes.
                                Defaults to the set defined in this module.
        :type file_attributes: dict

        :param var_attributes: dist of variable names, and the keys and values for variable
                               attributes.
                               Defaults to the set defined in this module.
        :type var_attributes: dict

        :param nc_version=3: version of netcdf to use -- must be 3 or 4. If 4, some extra
                             features are enabled.
        :type nc_version: integer
        """

        self.num_timesteps = num_timesteps
        self.ref_time = ref_time

        self.file_attributes = file_attributes
        self.var_attributes = var_attributes
        self.nc = None

        try:
            nc_version = int(nc_version)
        except ValueError:
            raise ValueError("nc_format must be 3 or 4")
        if nc_version == 3:
            format = 'NETCDF3_CLASSIC'
        elif nc_version == 4:
            format = 'NETCDF4'
        else:
            raise ValueError("nc_format must be 3 or 4")

        if nc_version == 3 and self.num_timesteps is None:
            raise ValueError("You must specify num_timesteps when using netcdf3")

        nc = netCDF4.Dataset(filename, 'w', format=format)
        self.nc = nc

        # Global attributes
        for (name, value) in self.file_attributes.items():
            setattr(nc, name, value)

        # Dimensions
        nc.createDimension('time', self.num_timesteps)
        nc.createDimension('data', None)

        # required variables
        time = nc.createVariable('time', np.int32, ('time',))
        for name, value in self.var_attributes['time'].items():
            time.setncattr(name, value)
        # make sure there are some units there
        # this will get overwritten when the proper reference time is known
        if self.ref_time is None:
            # make sure there are some units there
            # this will get overwritten when the proper reference time is known
            time.units = "seconds since 2016-01-01T00:00:00"
        else:
            time.units = 'seconds since {0}'.format(self.ref_time.isoformat())

        pc = nc.createVariable('particle_count', np.int32, ('time',))
        for name, value in self.var_attributes['particle_count'].items():
            pc.setncattr(name, value)
        self.time_var = time

        self.num_data = 0
        self.current_timestep = 0

    def write_timestep(self, timestamp, data):
        """
        write the data for a timestep

        :param timestamp: the time stamp of the timestep
        :type timestamp: datetime object

        :param data: dict of data arrays -- all parameters for a single time step
        :type data: dict

        Note: it is assumed that the timesteps will be written sequentially,
              and that the variables will not change after the first timestep
              is written.
        """

        nc = self.nc
        if self.current_timestep == 0:
            # create the variables and add attributes
            # set the time units:
            if self.ref_time is None:
                self.ref_time = timestamp
            nc.variables['time'].units = 'seconds since {0}'.format(self.ref_time.isoformat())
            for key, val in data.iteritems():
                val = np.asarray(val)
                var = nc.createVariable(key, datatype=val.dtype, dimensions=('data'))
                # if it's a standard variable, add the attributes
                if key in self.var_attributes:
                    for name, value in self.var_attributes[key].items():
                        var.setncattr(name, value)

        particle_count = len(data.itervalues().next())  # length of an arbitrary array
        nc.variables['particle_count'][self.current_timestep] = particle_count
        nc.variables['time'][self.current_timestep] = (timestamp - self.ref_time).total_seconds()
        self.current_timestep += 1
        for key, val in data.iteritems():
            var = nc.variables[key]
            if len(val) != particle_count:
                raise ValueError("All data arrays must be the same length")
            var[self.num_data:] = val
        self.num_data += particle_count

    def close(self):
        """
        close the netcdf file
        """
        if self.nc is not None:
            try:
                self.nc.close()
            except RuntimeError:
                # just in case it isn't still open
                pass

    def __del__(self):
        """ make sure to close the netcdf file """
        self.close()


class Reader(object):
    """
    Class to handle reading a nc_particle file

    (such as those written by GNOME or the Writer class above)
    """
    def __init__(self, nc_file):
        """
        initialize a file reader.

        :param nc_file: the netcdf file to read. If a netCDF4 Dataset, it will be used,
                        if a string, a new netCDF Dataset will be opened for reading
                        using that filename
        :type nc_file: string or netCDF4 Dataset object

        """

        if type(nc_file) == netCDF4.Dataset:
            # already open -- just use it
            self.nc = nc_file
            self.manage_dataset = False
        else:
            # open a new one
            self.nc = netCDF4.Dataset(nc_file)
            self.manage_dataset = True

        time = self.nc.variables['time']
        units = time.getncattr('units')
        self.times = netCDF4.num2date(time[:], units)
        self.time_units = units

        self.particle_count = self.nc.variables['particle_count']
        # build the index:
        self.data_index = np.zeros((len(self.times) + 1,), dtype=np.int32)
        self.data_index[1:] = np.cumsum(self.particle_count)

    @property
    def variables(self):
        """
        return the names of all the variables associated with the particles
        """
        return [var for var in self.nc.variables.keys() if var not in SPECIAL_VARIABLES]

    def __str__(self):
        return ("nc_particles Reader object:\n"
                "variables: {}\n"
                "number of timesteps: {}\n"
                ).format(self.variables, len(self.times))

    def get_all_timesteps(self, variables=['latitude', 'longitude']):
        """
        returns the requested variables data from all timesteps as a
        dictionary keyed by the variable names

        :param variables: the variables desired as a list string names.
                          Defaults to ['latitude','longitude']
        :type variables: list of strings

        :returns data: returns a dict of arrays -- the keys are the
                       variable names, and the values are numpy arrays
                        of the data. The arrays are the flattened ragged
                        array of data.
        """
        data = {}
        for var in variables:
            data[var] = []
            for i in range(len(self.times)):
                ind1 = self.data_index[i]
                ind2 = self.data_index[i + 1]
                data[var].append(self.nc.variables[var][ind1:ind2])
        return data

    def get_units(self, variable):
        """
        return the units of the given variable

        :param variable: name of the variable for which the units are required
        :type variable: string
        """
        return self.nc.variables[variable].units

    def get_attributes(self, variable):
        """
        return all the attributes of the given variable

        :param variable: name of the variable for which the attributes are required
        :type variable: string
        """
        var = self.nc.variables[variable]
        return {name: var.getncattr(name) for name in var.ncattrs()}

    def get_timestep(self, timestep, variables=['latitude', 'longitude']):
        """
        returns the requested variables data from a given timestep as a
        dictionary keyed by the variable names

        :param timestep: Index of the timestep you want the data for.
        :type timestep: int

        :param variables: The variables desired as a list string names.
                          Defaults to ['latitude','longitude']
        :type variables: list of strings

        :returns data: returns a dict of arrays -- the keys are the
                       variable names, and the values are numpy arrays
                       of the data.
        """
        ind1, ind2 = self.data_index[timestep:timestep + 2]
        return {var: self.nc.variables[var][ind1:ind2] for var in variables}

    def get_individual_trajectory(self, particle_id, variables=['latitude', 'longitude']):
        """
        returns the requested variables from trajectory of an individual particle

        :param particle_id: the id of the particle you want to track

        :param variables=['latitude', 'longitude']: which variables you want

        note: this is inefficient -- it has to read the entire file to get it.
        """
        indexes = np.where(self.nc.variables['id'][:] == particle_id)
        data = {}
        for var in variables:
            data[var] = self.nc.variables[var][indexes]
        return data

    def close(self):
        """
        close the netcdf file
        """
        if self.nc.isopen():
            self.nc.close()


    def __del__(self):
        """
        Make sure to close the netcdf file

        But only if this instance opened it in the first place
        """

        if self.manage_dataset:
            self.close()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Feb 2012
#
#------------------------------------------------------------------------------

"""
Test that menu and toolbar actions are triggered.
"""
import pyface

from traits.has_traits import HasTraits
from traits.trait_types import Bool
from traitsui.menu import Action, ActionGroup, Menu, MenuBar, ToolBar
from traitsui.item import Item
from traitsui.view import View

from traitsui.tests._tools import *
from traitsui.tests._tools import _is_current_backend

if _is_current_backend('null'):
    import nose
    raise nose.SkipTest("Not supported using the null backend")


TestAction = Action(
    name='Test',
    action='test_clicked',
    tooltip='Click to test'
)


class DialogWithToolbar(HasTraits):
    """Test dialog with toolbar and menu."""

    action_successful = Bool(False)

    def test_clicked(self):
        self.action_successful = True

    menubar = MenuBar(
        Menu(
            ActionGroup(TestAction),
            name='&Test menu'
        ),
    )

    toolbar = ToolBar(
        ActionGroup(TestAction),
    )

    traits_view = View(
        Item(label="Click the button on the toolbar or the menu item.\n"
                   "The 'Action successful' element should turn to True."),
        Item('action_successful', style='readonly'),
        menubar=menubar,
        toolbar=toolbar,
        buttons=['OK']
    )


def _test_actions(trigger_action_func):
    """Template test for wx, qt4, menu, and toolbar testing.
    """
    # Behavior: when clicking on a menu or toolbar action,
    # the corresponding function should be executed

    with store_exceptions_on_all_threads():
        # create dialog with toolbar adn menu
        dialog = DialogWithToolbar()
        ui = dialog.edit_traits()

        # press toolbar or menu button
        trigger_action_func(ui)

        # verify that the action was triggered
        nose.tools.assert_true(dialog.action_successful)


# ----- Qt4 tests

def _qt_trigger_action(container_class, ui):
    toolbar = ui.control.findChild(container_class)
    action = toolbar.actions()[0]
    action.trigger()


@skip_if_not_qt4
def test_qt_toolbar_action():
    # Behavior: when clicking on a toolbar action, the corresponding function
    # should be executed

    # Bug: in the Qt4 backend, a
    # TypeError: perform() takes exactly 2 arguments (1 given) was raised
    # instead

    qt_trigger_toolbar_action = partial(
        _qt_trigger_action, pyface.ui.qt4.action.tool_bar_manager._ToolBar)

    _test_actions(qt_trigger_toolbar_action)


@skip_if_not_qt4
def test_qt_menu_action():
    # Behavior: when clicking on a menu action, the corresponding function
    # should be executed

    # Bug: in the Qt4 backend, a
    # TypeError: perform() takes exactly 2 arguments (1 given) was raised
    # instead

    qt_trigger_menu_action = partial(
        _qt_trigger_action, pyface.ui.qt4.action.menu_manager._Menu)

    _test_actions(qt_trigger_menu_action)


# ----- wx tests

@skip_if_not_wx
def test_wx_toolbar_action():
    # Behavior: when clicking on a toolbar action, the corresponding function
    # should be executed

    import wx

    def _wx_trigger_toolbar_action(ui):
        # long road to get at the Id of the toolbar button
        toolbar_item = ui.view.toolbar.groups[0].items[0]
        toolbar_item_wrapper = toolbar_item._wrappers[0]
        control_id = toolbar_item_wrapper.control_id

        # build event that clicks the button
        click_event = wx.CommandEvent(wx.wxEVT_COMMAND_TOOL_CLICKED,
                                      control_id)

        # send the event to the toolbar
        toolbar = ui.control.FindWindowByName('toolbar')
        toolbar.ProcessEvent(click_event)

    _test_actions(_wx_trigger_toolbar_action)


# TODO: I couldn't find a way to press menu items programmatically for wx


if __name__ == '__main__':
    # Execute from command line for manual testing
    vw = DialogWithToolbar()
    vw.configure_traits()
